#ifndef _LIB_SYSTEM_H_
#define _LIB_SYSTEM_H_
#include <arpa/inet.h>
#include "platform.h"
#include "portable.h"

/*======================Macro Descript================================*/
#define SYS_USER_LEN                21
#define SYS_PASSWD_LEN              11
#define SYS_CRYPT_LEN               14
#define CUI_USER_LEN		    SYS_USER_LEN
#define CUI_PASS_LEN		    SYS_CRYPT_LEN

#define System_OK                   0
#define System_ReadConfErr          -1
#define System_WriteConfErr         -2


#define WAN_GetStatusOK             0
#define WAN_GetStatusErr            -1

#define ETH_ALEN                    6
#define DEVICE_NAME_LEN             20
#define FW_VERSION_LEN              40
#define DESCRIPT_LEN                50

#define PASSWD_FILE                 "/var/.htpasswd"
#define MAC_PREFIX                  "VR"

enum{
    ADD_IN_HEAD,
    ADD_IN_TAIL
};

#define RESOLV_FILE     "/etc/resolv.conf"
#define MAX_DNS_NUM                 4

#define ETH_WAN         "eth0"
#define PPP0_WAN	"ppp0"
#define DOD_IP "10.64.64.64"
       
/*======================Struct Descript================================*/

typedef struct{
    char UserName[SYS_USER_LEN];        /* End with null */
    char Password[SYS_CRYPT_LEN];       /* End with null */
    char Connection;
    char cuiUserName[CUI_USER_LEN];
    char cuiPassword[CUI_PASS_LEN];
}SYSTEM_CONF;

extern SYSTEM_CONF system_conf;

enum{
    SYS_ITEM_USERNAME,
    SYS_ITEM_PASSWD,
    SYS_ITEM_CONNECTION,
    SYS_ITEM_CUIUSERNAME,
    SYS_ITEM_CUIPASSWD,
    SYS_ITEM_NUMBER
    
};

enum{
    CONNECTION_DHCP,
    CONNECTION_SINGLE_PPPOE,
    CONNECTION_MULTI_PPPOE,
#ifdef _WAN_MODEM_
    CONNECTION_MODEM,
#endif
	CONNECTION_LOCAL_OFFICE,
	CONNECTION_NUM,
};

enum{
    CONNECTION_OK,
    CONNECTION_FAIL,
    CONNECTION_ERROR,
};


typedef struct{
    BYTE mac[6];
    INT16 flag;
    struct in_addr ip;
    struct in_addr mask;
    struct in_addr gw;    
}IFACE_STATUS;

extern IFACE_STATUS if_status;


typedef struct S_HWID {
#ifdef _BIG_ENDIAN_
  /*----- Interface --------------------------*/
  BYTE  fPrn      :2;  /* Printer */
  BYTE  fMIO      :1;  /* MIO */
  BYTE  fChinese  :1;  /* Chinese */
  BYTE  fSCSI     :2;  /* SCSI */
  BYTE  fPrnIn    :2;  /* Printer port in */

  BYTE  fNetworkNo:4;  /* Number of network */
  BYTE  fSerialNo :4;  /* Number of serial port */

  /*----- System -----------------------------*/
  BYTE  fRomType  :1;  /* 0= Fladsh, 1= OTROM */
  BYTE  fRAM      :3;  /* 000= 32K, 001= 128K, 010= 256K, 011= 1024K */
  BYTE  fNVRAM    :2;  /* 00= 256B, 01= 512B, 10= 8192B */
  BYTE  fRsv_2    :2;
  BYTE  fCPU      :3;  /* 000= 188, 001= 186, 010= 386DX */

  BYTE  fSpeed    :2;  /*      188/186   386DX
                           00   12MHz    16MHz
                           01   16MHz    25MHz
                           10   25MHz    33MHz
                           11   40MHz    40MHz  */
  BYTE  fRomSize  :3;  /* 000= 64K, 001= 128K, 010= 256K, 011= 512K */

  /*----- Card number ------------------------*/
  BYTE  bVersion  :4;
  BYTE  bRsv_1    :4;

  BYTE  fProduct2 :4;
  BYTE  fDetail   :4;

  BYTE  fProduct0 :4;
  BYTE  fProduct1 :4;
#else /* _BIG_ENDIAN_ */
  /*----- Card number ------------------------*/
  BYTE  fProduct1 :4;
  BYTE  fProduct0 :4;

  BYTE  fDetail   :4;
  BYTE  fProduct2 :4;

  BYTE  bRsv_1    :4;
  BYTE  bVersion  :4;

  /*----- System -----------------------------*/
  BYTE  fRomSize  :3;  /* 000= 64K, 001= 128K, 010= 256K, 011= 512K */
  BYTE  fSpeed    :2;  /*      188/186   386DX
                           00   12MHz    16MHz
                           01   16MHz    25MHz
                           10   25MHz    33MHz
                           11   40MHz    40MHz  */
  BYTE  fCPU      :3;  /* 000= 188, 001= 186, 010= 386DX */

  BYTE  fRsv_2    :2;
  BYTE  fNVRAM    :2;  /* 00= 256B, 01= 512B, 10= 8192B */
  BYTE  fRAM      :3;  /* 000= 32K, 001= 128K, 010= 256K, 011= 1024K */
  BYTE  fRomType  :1;  /* 0= Fladsh, 1= OTROM */

  /*----- Interface --------------------------*/
  BYTE  fSerialNo :4;  /* Number of serial port */
  BYTE  fNetworkNo:4;  /* Number of network */

  BYTE  fPrnIn    :2;  /* Printer port in */
  BYTE  fSCSI     :2;  /* SCSI */
  BYTE  fChinese  :1;  /* Chinese */
  BYTE  fMIO      :1;  /* MIO */
  BYTE  fPrn      :2;  /* Printer */
#endif /* _BIG_ENDIAN_ */

  WORD  wRsv_3;

  /*----- description ------------------------*/
  BYTE  bComment[23];

} PACKED HWID;


typedef struct S_PID {
  /*----- PID version ------------------------*/
  WORD  wPidVersion;        /* Motorola format */

  /*----- D/L control ------------------------*/
  BYTE  fReserved;

#ifdef _BIG_ENDIAN_
  BYTE  fRsv        :  4;
  BYTE  fChkFunction:  1;
  BYTE  fChkProtocol:  1;
  BYTE  fChkProduct:   1;
  BYTE  fChkHw:        1;
#else /* _BIG_ENDIAN_ */
  BYTE  fChkHw:        1;
  BYTE  fChkProduct:   1;
  BYTE  fChkProtocol:  1;
  BYTE  fChkFunction:  1;
  BYTE  fRsv        :  4;
#endif /* _BIG_ENDIAN_ */

  /*----- Hardware ID ------------------------*/
  HWID  hw;

  /*----- Reserver ---------------------------*/
  WORD  wRsv;

  /*----- Product ID & mask ------------------*/
  WORD  wProductID;         /* Motorola format */
  WORD  wProductMask;       /* Motorola format */

  /*----- Protocol ID & mask -----------------*/
  WORD  wProtocolID;        /* Motorola format */
  WORD  wProtocolMask;      /* Motorola format */

  /*----- Function ID & mask -----------------*/
  WORD  wFunctionID;        /* Motorola format */
  WORD  wFunctionMask;      /* Motorola format */

  /*----- F/W version ------------------------*/
  WORD  wFwVersion;         /* Motorola format */

  /*----- Code address -----------------------*/
  WORD  wSegment;           /* Intel format */
  WORD  wOffset;            /* Intel format */

} PACKED PID;


typedef  struct  VCI  {
    WORD    VerControl;
    WORD    DownControl;
    BYTE    Hid[32];
    WORD    Hver;
    WORD    ProdID;
    WORD    ProdIDmask;
    WORD    ProtID;
    WORD    ProtIDmask;
    WORD    FuncID;
    WORD    FuncIDmask;
    WORD    Fver;
    WORD    Cseg;
	WORD    Csize;
} PACKED VCI_TABLE;

#define  VCI_LEN         56

typedef struct {
    unsigned char dbHWID[ETH_ALEN];
    char device_name[DEVICE_NAME_LEN];
    char FWVersion[FW_VERSION_LEN];
    char FullFWVersion[FW_VERSION_LEN];
    char descript[DESCRIPT_LEN];
}DEVICE_INFO;

/*======================Function Descript================================*/

/*
 * crypt_password()   This function will call crypt() to encrypt psw_org(User input)
 *                    return value will be 13 printable ASCII characters.
 */
char *crypt_password(const char* psw_org);


int ReadSystemSettings(SYSTEM_CONF *pConf);
int SaveSystemSettings(SYSTEM_CONF *pConf);

int CheckPasswd(const char *pass,const char*crypt_pass);
int ChangePasswd(SYSTEM_CONF *pConf);
int MakeHttpdPasswd(void);

int GetIfaceStatus(char *if_name, IFACE_STATUS *pStatus);
int GetConnectionStatus(IFACE_STATUS *pStatus);

void connectPPPoE(int interface);
void disconnectPPPoE(int interface);

PID *ReadProgramID(void);
int get_device_info(DEVICE_INFO *device);
void doSomethingBeforeReboot(void);
void do_reboot(void);
int restore(void);
void FactoryRestore(void);
struct in_addr *GetDNSIPFromFile(char *filename);

int GetPreferDNS(struct in_addr *dns_p);

int get_wan_type();
int set_wan_type(int wan_type);
#endif /* _SYSTEM_H_ */
