/* Driver for Philips webcam
   Functions that send various control messages to the webcam, including
   video modes.
   (C) 1999-2001 Nemosoft Unv. (webcam@smcc.demon.nl)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
   Changes
   2001/08/03  Alvarado   Added methods for changing white balance and
                          red/green gains
 */

/* Control functions for the cam; brightness, contrast, video mode, etc. */

/* =================================================================================
	03/03/31 [JTOPRO's camera control command
   ================================================================================= */

#ifdef __KERNEL__
//#include <asm/uaccess.h>
#endif
#include <asm/errno.h>
#include <linux/usb.h>
#include <stdio.h>
//#include <tcl.h>
#include "pwc.h"
#include "pwc-ioctl.h"
#include "pwc-uncompress.h"


// topro header
#include "tp_def.h"

#include "tp_gamma_r.h"
#include "tp_gamma_g.h"
#include "tp_gamma_b.h"

#include "tp_param.h"
#include "lnx.h"
/* Request types: video */

#define SET_LUM_CTL			0x01
#define GET_LUM_CTL			0x02
#define SET_CHROM_CTL			0x03
#define GET_CHROM_CTL			0x04
#define SET_STATUS_CTL			0x05
#define GET_STATUS_CTL			0x06
#define SET_EP_STREAM_CTL		0x07
#define GET_EP_STREAM_CTL		0x08

/* Selectors for the Luminance controls [G/S]ET_LUM_CTL */
#define AGC_MODE_FORMATTER			0x2000
#define PRESET_AGC_FORMATTER			0x2100
#define SHUTTER_MODE_FORMATTER			0x2200
#define PRESET_SHUTTER_FORMATTER		0x2300
#define PRESET_CONTOUR_FORMATTER		0x2400
#define AUTO_CONTOUR_FORMATTER			0x2500
#define BACK_LIGHT_COMPENSATION_FORMATTER	0x2600
#define CONTRAST_FORMATTER			0x2700
#define DYNAMIC_NOISE_CONTROL_FORMATTER		0x2800
#define FLICKERLESS_MODE_FORMATTER		0x2900
#define AE_CONTROL_SPEED			0x2A00
#define BRIGHTNESS_FORMATTER			0x2B00
#define GAMMA_FORMATTER				0x2C00

/* Selectors for the Chrominance controls [G/S]ET_CHROM_CTL */
#define WB_MODE_FORMATTER			0x1000
#define AWB_CONTROL_SPEED_FORMATTER		0x1100
#define AWB_CONTROL_DELAY_FORMATTER		0x1200
#define PRESET_MANUAL_RED_GAIN_FORMATTER	0x1300
#define PRESET_MANUAL_BLUE_GAIN_FORMATTER	0x1400
#define COLOUR_MODE_FORMATTER			0x1500
#define SATURATION_MODE_FORMATTER1		0x1600
#define SATURATION_MODE_FORMATTER2		0x1700

/* Selectors for the Status controls [G/S]ET_STATUS_CTL */
#define SAVE_USER_DEFAULTS_FORMATTER		0x0200
#define RESTORE_USER_DEFAULTS_FORMATTER		0x0300
#define RESTORE_FACTORY_DEFAULTS_FORMATTER	0x0400
#define READ_AGC_FORMATTER			0x0500
#define READ_SHUTTER_FORMATTER			0x0600
#define READ_RED_GAIN_FORMATTER			0x0700
#define READ_BLUE_GAIN_FORMATTER		0x0800
#define READ_RAW_Y_MEAN_FORMATTER		0x3100
#define SET_POWER_SAVE_MODE_FORMATTER		0x3200
#define MIRROR_IMAGE_FORMATTER			0x3300
#define LED_FORMATTER				0x3400
#define VGA             1
#define QVGA            2



/* Formatters for the Video Endpoint controls [G/S]ET_EP_STREAM_CTL */
#define VIDEO_OUTPUT_CONTROL_FORMATTER		0x0100



#if TP6800_HV7131R
//hynix

#define GAMMA_R_GAIN      307
#define GAMMA_G_GAIN      258
#define GAMMA_B_GAIN      276
#define GAMMA_CONTRAST      11
#define GAMMA_BRIGHTNESS  -41
#define DEFAULT_EXPOSURE    20

#elif  TP6800_MT9V011

#define GAMMA_R_GAIN      315
#define GAMMA_G_GAIN      212
#define GAMMA_B_GAIN      255
#define GAMMA_CONTRAST      21
#define GAMMA_BRIGHTNESS  -25
#define DEFAULT_EXPOSURE    10

#elif  TP6800_SHARPCCD

#define GAMMA_R_GAIN      247
#define GAMMA_G_GAIN      212
#define GAMMA_B_GAIN      235
#define GAMMA_CONTRAST      11
#define GAMMA_BRIGHTNESS  -25
#define DEFAULT_EXPOSURE    30

#elif TP6800_EVS

#define GAMMA_R_GAIN      292
#define GAMMA_G_GAIN      258
#define GAMMA_B_GAIN      272
#define GAMMA_CONTRAST     13
#define GAMMA_BRIGHTNESS  -71
#define DEFAULT_EXPOSURE    30

#elif TP6800_TAS

#define GAMMA_R_GAIN      292
#define GAMMA_G_GAIN      258
#define GAMMA_B_GAIN      272
#define GAMMA_CONTRAST     13
#define GAMMA_BRIGHTNESS  -71
#define DEFAULT_EXPOSURE    30


#elif TP6800_PAS302

#define GAMMA_R_GAIN      265
#define GAMMA_G_GAIN      242
#define GAMMA_B_GAIN      259
#define GAMMA_CONTRAST      28
#define GAMMA_BRIGHTNESS  -110
#define DEFAULT_EXPOSURE    15


#endif


unsigned char GlobalGain=0x50;
int RFactor=50, GFactor=50, BFactor=50;
int topro_brightness=60;  //0~100
int topro_contrast=60;    //0~100
int topro_gamma=10;       //0~30
int AutoExposure=0,AutoWhiteBalance=0;


static char *size2name[PSZ_MAX] =
{
	"subQCIF",
	"QSIF",
	"QCIF",
	"SIF",
	"CIF",
	"VGA",
};  





/********/

/* Entries for the Nala (645/646) camera; the Nala doesn't have compression 
   preferences, so you either get compressed or non-compressed streams.
   
   An alternate value of 0 means this mode is not available at all.
 */

struct Nala_table_entry {
	char alternate;			/* USB alternate setting */
	int compressed;			/* Compressed yes/no */

	unsigned char mode[3];		/* precomputed mode table */
};

static struct Nala_table_entry Nala_table[PSZ_MAX][8] =
{

#include "pwc_nala.h"
};

/* This tables contains entries for the 675/680/690 (Timon) camera, with
   4 different qualities (no compression, low, medium, high).
   It lists the bandwidth requirements for said mode by its alternate interface 
   number. An alternate of 0 means that the mode is unavailable.
   
   There are 6 * 4 * 4 entries: 
     6 different resolutions subqcif, qsif, qcif, sif, cif, vga
     6 framerates: 5, 10, 15, 20, 25, 30
     4 compression modi: none, low, medium, high
     
   When an uncompressed mode is not available, the next available compressed mode 
   will be choosen (unless the decompressor is absent). Sometimes there are only
   1 or 2 compressed modes available; in that case entries are duplicated.
*/
struct Timon_table_entry 
{
	char alternate;			/* USB alternate interface */
	unsigned short packetsize;	/* Normal packet size */
	unsigned short bandlength;	/* Bandlength when decompressing */
	unsigned char mode[13];		/* precomputed mode settings for cam */
};

static struct Timon_table_entry Timon_table[PSZ_MAX][6][4] = 
{
#include "pwc_timon.h"
};

/* Entries for the Kiara (730/740/750) camera */

struct Kiara_table_entry
{
	char alternate;			/* USB alternate interface */
	unsigned short packetsize;	/* Normal packet size */
	unsigned short bandlength;	/* Bandlength when decompressing */
	unsigned char mode[12];		/* precomputed mode settings for cam */
};

static struct Kiara_table_entry Kiara_table[PSZ_MAX][6][4] =
{
#include "pwc_kiara.h"
};


/****************************************************************************/




#if PWC_DEBUG
void pwc_hexdump(void *p, int len)
{
	int i;
	unsigned char *s;
	char buf[100], *d;

	s = (unsigned char *)p;
	d = buf;
	*d = '\0';
	Debug("Doing hexdump @ %p, %d bytes.\n", p, len);
	for (i = 0; i < len; i++) {
		d += sprintf(d, "%02X ", *s++);
		if ((i & 0xF) == 0xF) {
			Debug("%s\n", buf);
			d = buf;
			*d = '\0';
		}
	}
	if ((i & 0xF) != 0)
		Debug("%s\n", buf);
}

#endif

static inline int send_video_command(struct usb_device *udev, int index, void *buf, int buflen)
{
	int i;

	    i = usb_control_msg(udev,
		usb_sndctrlpipe(udev, 0),
		SET_EP_STREAM_CTL,
		USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		VIDEO_OUTPUT_CONTROL_FORMATTER,
		index,
		buf, buflen, HZ);

	printk("send_cmd return %d/%08x\n",i,usb_sndctrlpipe(udev, 0));

	return i;
}



static inline int set_video_mode_Nala(struct pwc_device *pdev, int size, int frames)
{
	unsigned char buf[3];
	int ret, fps;
	struct Nala_table_entry *pEntry;
	int frames2frames[31] =
	{ /* closest match of framerate */
	   0,  0,  0,  0,  4,  /*  0-4  */
	   5,  5,  7,  7, 10,  /*  5-9  */
          10, 10, 12, 12, 15,  /* 10-14 */
          15, 15, 15, 20, 20,  /* 15-19 */
          20, 20, 20, 24, 24,  /* 20-24 */
          24, 24, 24, 24, 24,  /* 25-29 */
          24                   /* 30    */
	};
	int frames2table[31] =
	{ 0, 0, 0, 0, 0, /*  0-4  */
	  1, 1, 1, 2, 2, /*  5-9  */
	  3, 3, 4, 4, 4, /* 10-14 */
	  5, 5, 5, 5, 5, /* 15-19 */
	  6, 6, 6, 6, 7, /* 20-24 */
	  7, 7, 7, 7, 7, /* 25-29 */
	  7              /* 30    */
	};
	
	if (size < 0 || size > PSZ_CIF || frames < 4 || frames > 25)
		return -EINVAL;
	frames = frames2frames[frames];
	fps = frames2table[frames];
	pEntry = &Nala_table[size][fps];
	if (pEntry->alternate == 0)
		return -EINVAL;

	if (pEntry->compressed && pdev->decompressor == NULL)
		return -ENOENT; /* Not supported. */

	memcpy(buf, pEntry->mode, 3);	
	ret = send_video_command(pdev->udev, pdev->vendpoint, buf, 3);
	if (ret < 0)
		return ret;


	if (pEntry->compressed)
		pdev->decompressor->init(pdev->release, buf, pdev->decompress_data);
		
	/* Set various parameters */
	pdev->vframes = frames;
	pdev->vsize = size;
	pdev->valternate = pEntry->alternate;
	pdev->image = pwc_image_sizes[size];
	pdev->frame_size = (pdev->image.x * pdev->image.y * 3) / 2;
	if (pEntry->compressed) {
		if (pdev->release < 5) { /* 4 fold compression */
			pdev->vbandlength = 528;
			pdev->frame_size /= 4;
		}
		else {
			pdev->vbandlength = 704;
			pdev->frame_size /= 3;
		}
	}
	else
		pdev->vbandlength = 0;
	return 0;
}


static inline int set_video_mode_Timon(struct pwc_device *pdev, int size, int frames, int compression, int snapshot)
{
	unsigned char buf[13];
	struct Timon_table_entry *pChoose;
	int ret, fps;

	if (size >= PSZ_MAX || frames < 5 || frames > 30 || compression < 0 || compression > 3)
		return -EINVAL;
	if (size == PSZ_VGA && frames > 15)
		return -EINVAL;
	fps = (frames / 5) - 1;

	/* Find a supported framerate with progressively higher compression ratios
	   if the preferred ratio is not available.
	*/
	pChoose = NULL;
	if (pdev->decompressor == NULL) {
#if PWC_DEBUG	
		Debug("Trying to find uncompressed mode.\n");
#endif
		pChoose = &Timon_table[size][fps][0];
	}
	else {
		while (compression <= 3) {
			pChoose = &Timon_table[size][fps][compression];
			if (pChoose->alternate != 0)

				break;
			compression++;	
		}
	}
	if (pChoose == NULL || pChoose->alternate == 0)
		return -ENOENT; /* Not supported. */

	memcpy(buf, pChoose->mode, 13);
	if (snapshot)
		buf[0] |= 0x80;
	if ( pdev->type != 800 ) {
	ret = send_video_command(pdev->udev, pdev->vendpoint, buf, 13);
	if (ret < 0)
		return ret;
	}

	if (pChoose->bandlength > 0) {
		printk("decompressor init\n");
		pdev->decompressor->init(pdev->release, buf, pdev->decompress_data);
	}
	/* Set various parameters */
	pdev->vframes = frames;
	pdev->vsize = size;
	pdev->vsnapshot = snapshot;
	if ( pdev->type != 800 ) pdev->valternate = pChoose->alternate;
	else pdev->valternate = 6;
	pdev->image = pwc_image_sizes[size];
	pdev->vbandlength = pChoose->bandlength;
	//rintk("bandlength = %d\n",pdev->vbandlength);
	if (pChoose->bandlength > 0) 
		if ( pdev->type == 800 ) pdev->frame_size = 40960;
		else pdev->frame_size = (pChoose->bandlength * pdev->image.y) / 4;
	else
		if ( pdev->type == 800 ) pdev->frame_size = 40960;
		else pdev->frame_size = (pdev->image.x * pdev->image.y * 12) / 8;
	
	return 0;
}


static inline int set_video_mode_Kiara(struct pwc_device *pdev, int size, int frames, int compression, int snapshot)
{
	struct Kiara_table_entry *pChoose;
	int fps, ret;
	unsigned char buf[12];
	
	if (size >= PSZ_MAX || frames < 5 || frames > 30 || compression < 0 || compression > 3)
		return -EINVAL;
	if (size == PSZ_VGA && frames > 15)
		return -EINVAL;

	fps = (frames / 5) - 1;
	
	/* Find a supported framerate with progressively higher compression ratios

	   if the preferred ratio is not available.
	*/
	pChoose = NULL;
	if (pdev->decompressor == NULL) {
#if PWC_DEBUG
		Debug("Trying to find uncompressed mode.\n");
#endif		
		pChoose = &Kiara_table[size][fps][0];
	}
	else {
		while (compression <= 3) {
			pChoose = &Kiara_table[size][fps][compression];
			if (pChoose->alternate != 0)
				break;
			compression++;
		}
	}
	if (pChoose == NULL || pChoose->alternate == 0)
		return -ENOENT; /* Not supported. */

	/* usb_control_msg won't take staticly allocated arrays as argument?? */
	memcpy(buf, pChoose->mode, 12);
	if (snapshot)
		buf[0] |= 0x80;

	/* Firmware bug: video endpoint is 5, but commands are sent to endpoint 4 */
	ret = send_video_command(pdev->udev, 4 /* pdev->vendpoint */, buf, 12);
	if (ret < 0)
		return ret;

	if (pChoose->bandlength > 0)
		pdev->decompressor->init(pdev->release, buf, pdev->decompress_data);


	/* All set and go */
	pdev->vframes = frames;
	pdev->vsize = size;
	pdev->vsnapshot = snapshot;
	pdev->valternate = pChoose->alternate;
	pdev->image = pwc_image_sizes[size];
	pdev->vbandlength = pChoose->bandlength;
	if (pChoose->bandlength > 0)
		pdev->frame_size = (pChoose->bandlength * pdev->image.y) / 4;
	else 
		pdev->frame_size = (pdev->image.x * pdev->image.y * 12) / 8;
	pdev->frame_size += (pdev->frame_header_size + pdev->frame_trailer_size);
	return 0;
}


/**
   @pdev: device structure
   @width: viewport width
   @height: viewport height
   @frame: framerate, in fps
   @compression: preferred compression ratio
   @snapshot: snapshot mode or streaming
 */
int pwc_set_video_mode(struct pwc_device *pdev, int width, int height, int frames, int compression, int snapshot)


{
	int ret, size;
	
	size = pwc_decode_size(pdev, width, height);
	if (size < 0) {
		Debug("Could not find suitable size.\n");
		return -ERANGE;
	}
	ret = -EINVAL;	
	switch(pdev->type) {
	case 645:
	case 646:
		ret = set_video_mode_Nala(pdev, size, frames);
		break;

	case 675:
	case 680:
	case 690:
	case 800:
		ret = set_video_mode_Timon(pdev, size, frames, compression, snapshot);
		break;

	case 730:
	case 740:
	case 750:
		ret = set_video_mode_Kiara(pdev, size, frames, compression, snapshot);
		break;
	}
	if (ret < 0) {
		if (ret == -ENOENT)
			Info("Video mode %s@%d fps is only supported with the decompressor module (pwcx).\n", size2name[size], frames);
		else {
			Err("Failed to set video mode %s@%d fps; return code = %d\n", size2name[size], frames, ret);
			return ret;

		}
	}
	pdev->view.x = width;
	pdev->view.y = height;
	pwc_set_image_buffer_size(pdev);
	Trace(TRACE_SIZE, "Set viewport to %dx%d, image size is %dx%d, palette = %d.\n", width, height, pwc_image_sizes[size].x, pwc_image_sizes[size].y, pdev->vpalette);
	return 0;
}


void pwc_set_image_buffer_size(struct pwc_device *pdev)
{
	int factor, i, filler = 0;

	switch(pdev->vpalette) {
	case VIDEO_PALETTE_RGB32 | 0x80:
	case VIDEO_PALETTE_RGB32:
		factor = 16;
		filler = 0;
		break;
	case VIDEO_PALETTE_RGB24 | 0x80:

	case VIDEO_PALETTE_RGB24:

		factor = 12;
		filler = 0;
		break;
	case VIDEO_PALETTE_YUYV:
	case VIDEO_PALETTE_YUV422:
		factor = 8;
		filler = 128;
		break;
	case VIDEO_PALETTE_YUV420:
	case VIDEO_PALETTE_YUV420P:
		factor = 6;
		filler = 128;
		break;
#if PWC_DEBUG
	case VIDEO_PALETTE_RAW:
		pdev->image.size = pdev->frame_size;
		pdev->view.size = pdev->frame_size;
		return;
		break;
#endif	
	default:
		factor = 0;
		break;
	}

	/* Set sizes in bytes */
	pdev->image.size = pdev->image.x * pdev->image.y * factor / 4;
	pdev->view.size  = pdev->view.x  * pdev->view.y  * factor / 4;

	/* Align offset, or you'll get some very weird results in
	   YUV420 mode... x must be multiple of 4 (to get the Y's in 
	   place), and y even (or you'll mixup U & V). This is less of a
	   problem for YUV420P.
	 */
	pdev->offset.x = ((pdev->view.x - pdev->image.x) / 2) & 0xFFFC;
	pdev->offset.y = ((pdev->view.y - pdev->image.y) / 2) & 0xFFFE;
	
	/* Fill buffers with gray or black */
	for (i = 0; i < MAX_IMAGES; i++) {
		if (pdev->image_ptr[i] != NULL)
			memset(pdev->image_ptr[i], filler, pdev->view.size);
	}
}



/* BRIGHTNESS */

int pwc_get_brightness(struct pwc_device *pdev)
{
	char buf;
	int ret;
	
	ret = usb_control_msg(pdev->udev, usb_rcvctrlpipe(pdev->udev, 0),
		GET_LUM_CTL,
		USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		BRIGHTNESS_FORMATTER,
		pdev->vcinterface,
		&buf, 1, HZ / 2);
	if (ret < 0)
		return ret;
printk("pwc_get_brightness retunr %d/%x/%08x\n",ret,buf,usb_rcvctrlpipe(pdev->udev, 0));

	return buf << 9;
}

int pwc_set_brightness(struct pwc_device *pdev, int value)
{
	char buf;

	if (value < 0)
		value = 0;
	if (value > 0xffff)
		value = 0xffff;
	buf = (value >> 9) & 0x7f;
	return usb_control_msg(pdev->udev, usb_sndctrlpipe(pdev->udev, 0),
		SET_LUM_CTL,
		USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		BRIGHTNESS_FORMATTER,
		pdev->vcinterface,
		&buf, 1, HZ / 2);
}

/* CONTRAST */

int pwc_get_contrast(struct pwc_device *pdev)


{
	char buf;

	int ret;
	
	ret = usb_control_msg(pdev->udev, usb_rcvctrlpipe(pdev->udev, 0),
		GET_LUM_CTL,
		USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		CONTRAST_FORMATTER,
		pdev->vcinterface,
		&buf, 1, HZ / 2);
	if (ret < 0)
		return ret;
		

	printk("pwc_get_contrast ret %d/%x\n",ret,buf);
	return buf << 10;
}

int pwc_set_contrast(struct pwc_device *pdev, int value)
{
	char buf;

	if (value < 0)
		value = 0;
	if (value > 0xffff)
		value = 0xffff;
	buf = (value >> 10) & 0x3f;
	return usb_control_msg(pdev->udev, usb_sndctrlpipe(pdev->udev, 0),
		SET_LUM_CTL,
		USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		CONTRAST_FORMATTER,
		pdev->vcinterface,
		&buf, 1, HZ / 2);


}

/* GAMMA */

int pwc_get_gamma(struct pwc_device *pdev)
{
	char buf;
	int ret;

	ret = usb_control_msg(pdev->udev, usb_rcvctrlpipe(pdev->udev, 0),
		GET_LUM_CTL,
		USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		GAMMA_FORMATTER,

		pdev->vcinterface,

		&buf, 1, HZ / 2);
	if (ret < 0)
		return ret;
	return buf << 11;
}

int pwc_set_gamma(struct pwc_device *pdev, int value)
{
	char buf;



	if (value < 0)
		value = 0;
	if (value > 0xffff)
		value = 0xffff;
	buf = (value >> 11) & 0x1f;
	return usb_control_msg(pdev->udev, usb_sndctrlpipe(pdev->udev, 0),
		SET_LUM_CTL,
		USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		GAMMA_FORMATTER,
		pdev->vcinterface,
		&buf, 1, HZ / 2);
}


/* SATURATION */

int pwc_get_saturation(struct pwc_device *pdev)
{

	char buf;
	int ret;

	if (pdev->type < 675)
		return -1;
	ret = usb_control_msg(pdev->udev, usb_rcvctrlpipe(pdev->udev, 0),
		GET_CHROM_CTL,
		USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		pdev->type < 730 ? SATURATION_MODE_FORMATTER2 : SATURATION_MODE_FORMATTER1,

		pdev->vcinterface,
		&buf, 1, HZ / 2);
	if (ret < 0)
		return ret;
	return 32768 + buf * 327;
}

int pwc_set_saturation(struct pwc_device *pdev, int value)
{
	char buf;

	if (pdev->type < 675)
		return -EINVAL;

	if (value < 0)
		value = 0;
	if (value > 0xffff)
		value = 0xffff;
	/* saturation ranges from -100 to +100 */

	buf = (value - 32768) / 327;
	return usb_control_msg(pdev->udev, usb_sndctrlpipe(pdev->udev, 0),
		SET_CHROM_CTL,
		USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		pdev->type < 730 ? SATURATION_MODE_FORMATTER2 : SATURATION_MODE_FORMATTER1,


		pdev->vcinterface,
		&buf, 1, HZ / 2);
}

/* AGC */

static inline int pwc_set_agc(struct pwc_device *pdev, int mode, int value)

{
	char buf;
	int ret;

	if (mode)
		buf = 0x0; /* auto */
	else
		buf = 0xff; /* fixed */

	ret = usb_control_msg(pdev->udev, usb_sndctrlpipe(pdev->udev, 0),
		SET_LUM_CTL,
		USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		AGC_MODE_FORMATTER,
		pdev->vcinterface,
		&buf, 1, HZ / 2);
	
	if (!mode && ret >= 0) {
		if (value < 0)
			value = 0;
		if (value > 0xffff)
			value = 0xffff;
		buf = (value >> 10) & 0x3F;
		ret = usb_control_msg(pdev->udev, usb_sndctrlpipe(pdev->udev, 0),
			SET_LUM_CTL,
			USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
			PRESET_AGC_FORMATTER,
			pdev->vcinterface,

			&buf, 1, HZ / 2);
	}
	if (ret < 0)
		return ret;
	return 0;
}

static inline int pwc_get_agc(struct pwc_device *pdev, int *value)
{
	unsigned char buf;
	int ret;

	
	ret = usb_control_msg(pdev->udev, usb_rcvctrlpipe(pdev->udev, 0),
		GET_LUM_CTL,
		USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		AGC_MODE_FORMATTER,
		pdev->vcinterface,

		&buf, 1, HZ / 2);
	if (ret < 0)

		return ret;


	if (buf != 0) { /* fixed */
		ret = usb_control_msg(pdev->udev, usb_rcvctrlpipe(pdev->udev, 0),



			GET_LUM_CTL,
			USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
			PRESET_AGC_FORMATTER,
			pdev->vcinterface,
			&buf, 1, HZ / 2);
		if (ret < 0)
			return ret;
		if (buf > 0x3F)
			buf = 0x3F;
		*value = (buf << 10);		
	}
	else { /* auto */
		ret = usb_control_msg(pdev->udev, usb_rcvctrlpipe(pdev->udev, 0),
			GET_STATUS_CTL,
			USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
			READ_AGC_FORMATTER,
			pdev->vcinterface,
			&buf, 1, HZ / 2);
		if (ret < 0)
			return ret;

		/* Gah... this value ranges from 0x00 ... 0x9F */
		if (buf > 0x9F)
			buf = 0x9F;
		*value = -(48 + buf * 409);
	}

	return 0;
}

static inline int pwc_set_shutter_speed(struct pwc_device *pdev, int mode, int value)
{
	char buf[2];
	int speed, ret;


	if (mode)
		buf[0] = 0x0;	/* auto */
	else
		buf[0] = 0xff; /* fixed */
	
	ret = usb_control_msg(pdev->udev, usb_sndctrlpipe(pdev->udev, 0),

		SET_LUM_CTL,
		USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		SHUTTER_MODE_FORMATTER,
		pdev->vcinterface,
		buf, 1, HZ / 2);

	if (!mode && ret >= 0) {
		if (value < 0)
			value = 0;
		if (value > 0xffff)
			value = 0xffff;
		switch(pdev->type) {
		case 675:
		case 680:
		case 690:
			/* speed ranges from 0x0 to 0x290 (656) */
			speed = (value / 100);
			buf[1] = speed >> 8;
			buf[0] = speed & 0xff;
			break;
		case 730:
		case 740:

		case 750:
			/* speed seems to range from 0x0 to 0xff */
			buf[1] = 0;
			buf[0] = value >> 8;
			break;
		}


		ret = usb_control_msg(pdev->udev, usb_sndctrlpipe(pdev->udev, 0),
			SET_LUM_CTL,
			USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
			PRESET_SHUTTER_FORMATTER,
			pdev->vcinterface,
			&buf, 2, HZ / 2);
	}
	return ret;

}


/* POWER */

int pwc_camera_power(struct pwc_device *pdev, int power)
{
	char buf;

	if (pdev->type < 675 || (pdev->type < 730 && pdev->release < 6))
		return 0;	/* Not supported by Nala or Timon < release 6 */

	if (power)
		buf = 0x00; /* active */
	else
		buf = 0xFF; /* power save */
	return usb_control_msg(pdev->udev, usb_sndctrlpipe(pdev->udev, 0),
		SET_STATUS_CTL,
		USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		SET_POWER_SAVE_MODE_FORMATTER,

		pdev->vcinterface,
		&buf, 1, HZ / 2);
}



/* private calls */

static inline int pwc_restore_user(struct pwc_device *pdev)
{
	return usb_control_msg(pdev->udev, usb_sndctrlpipe(pdev->udev, 0),

		SET_STATUS_CTL,
		USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		RESTORE_USER_DEFAULTS_FORMATTER,


		pdev->vcinterface,
		NULL, 0, HZ / 2);
}

static inline int pwc_save_user(struct pwc_device *pdev)
{
	return usb_control_msg(pdev->udev, usb_sndctrlpipe(pdev->udev, 0),
		SET_STATUS_CTL,
		USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		SAVE_USER_DEFAULTS_FORMATTER,
		pdev->vcinterface,
		NULL, 0, HZ / 2);
}

static inline int pwc_restore_factory(struct pwc_device *pdev)
{

	return usb_control_msg(pdev->udev, usb_sndctrlpipe(pdev->udev, 0),
		SET_STATUS_CTL,
		USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		RESTORE_FACTORY_DEFAULTS_FORMATTER,
		pdev->vcinterface,
		NULL, 0, HZ / 2);
}

 /* ************************************************* */
 /* Patch by Alvarado: (not in the original version   */


 /*
  * the camera recognizes modes from 0 to 4:
  *
  * 00: indoor (incandescant lighting)
  * 01: outdoor (sunlight)
  * 02: fluorescent lighting
  * 03: manual
  * 04: auto
  */ 
static inline int pwc_set_awb(struct pwc_device *pdev, int mode)
{

	char buf;
	int ret;
	

	if (mode < 0)
	    mode = 0;
	
	if (mode > 4)
	    mode = 4;
	
	buf = mode & 0x07; /* just the lowest three bits */
	
	ret = usb_control_msg(pdev->udev, usb_sndctrlpipe(pdev->udev, 0),
		SET_CHROM_CTL,

		USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		WB_MODE_FORMATTER,
		pdev->vcinterface,
		&buf, 1, HZ / 2);

	if (ret < 0)
		return ret;
	return 0;
}


static inline int pwc_get_awb(struct pwc_device *pdev)
{
	unsigned char buf;
	int ret;
	
	ret = usb_control_msg(pdev->udev, usb_rcvctrlpipe(pdev->udev, 0),
		GET_CHROM_CTL,
		USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		WB_MODE_FORMATTER,
		pdev->vcinterface,
		&buf, 1, HZ / 2);

	if (ret < 0) 
		return ret;
	return buf;
}

static inline int pwc_set_red_gain(struct pwc_device *pdev, int value)
{
        unsigned char buf;

	if (value < 0)
		value = 0;
	if (value > 0xffff)
		value = 0xffff;

	/* only the msb are considered */
	buf = value >> 8;

	return usb_control_msg(pdev->udev, usb_sndctrlpipe(pdev->udev, 0),
		SET_CHROM_CTL,
		USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		PRESET_MANUAL_RED_GAIN_FORMATTER,
		pdev->vcinterface,
		&buf, 1, HZ / 2);
}

static inline int pwc_get_red_gain(struct pwc_device *pdev)
{
	unsigned char buf;
	int ret;
	
	ret = usb_control_msg(pdev->udev, usb_rcvctrlpipe(pdev->udev, 0),
 	        GET_STATUS_CTL,
		USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,

	        PRESET_MANUAL_RED_GAIN_FORMATTER,
		pdev->vcinterface,
		&buf, 1, HZ / 2);


	if (ret < 0)
	    return ret;
	
	return (buf << 8);
}


static inline int pwc_set_blue_gain(struct pwc_device *pdev, int value)
{
	unsigned char buf;

	if (value < 0)
		value = 0;
	if (value > 0xffff)
		value = 0xffff;

	/* linear mapping of 0..0xffff to -0x80..0x7f */
	buf = (value >> 8);

	return usb_control_msg(pdev->udev, usb_sndctrlpipe(pdev->udev, 0),
		SET_CHROM_CTL,
		USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		PRESET_MANUAL_BLUE_GAIN_FORMATTER,

		pdev->vcinterface,
		&buf, 1, HZ / 2);
}

static inline int pwc_get_blue_gain(struct pwc_device *pdev)
{
	unsigned char buf;
	int ret;

	printk("pwc_get_blue_gain rcvpipe = %d\n",usb_rcvctrlpipe(pdev->udev, 0));
	
	ret = usb_control_msg(pdev->udev, usb_rcvctrlpipe(pdev->udev, 0),
   	        GET_STATUS_CTL,
		USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		PRESET_MANUAL_BLUE_GAIN_FORMATTER,

		pdev->vcinterface,
		&buf, 1, HZ / 2);

	if (ret < 0)
	    return ret;
	
	printk("blue gain return %d\n",ret);
	
	return (buf << 8);
}

/* The following two functions are different, since they only read the
   internal red/blue gains, which may be different from the manual 
   gains set or read above.
 */   
static inline int pwc_read_red_gain(struct pwc_device *pdev)
{
	unsigned char buf;
	int ret;
	
	ret = usb_control_msg(pdev->udev, usb_rcvctrlpipe(pdev->udev, 0),
 	        GET_STATUS_CTL, 
		USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
	        READ_RED_GAIN_FORMATTER,

		pdev->vcinterface,
		&buf, 1, HZ / 2);

	if (ret < 0)
	    return ret;

	printk("red gain return %d\n",ret);
	
	return (buf << 8);
}

static inline int pwc_read_blue_gain(struct pwc_device *pdev)
{
	unsigned char buf;
	int ret;
	
	ret = usb_control_msg(pdev->udev, usb_rcvctrlpipe(pdev->udev, 0),
   	        GET_STATUS_CTL,
		USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		READ_BLUE_GAIN_FORMATTER,
		pdev->vcinterface,
		&buf, 1, HZ / 2);

	if (ret < 0)
	    return ret;

	
	return (buf << 8);
}

int pwc_set_leds(struct pwc_device *pdev, int on_value, int off_value)
{
	unsigned char buf[2];

	if (pdev->type < 730)
		return 0;
	on_value /= 100;
	off_value /= 100;
	if (on_value < 0)
		on_value = 0;
	if (on_value > 0xff)
		on_value = 0xff;

	if (off_value < 0)
		off_value = 0;
	if (off_value > 0xff)
		off_value = 0xff;

	buf[0] = on_value;
	buf[1] = off_value;

	return usb_control_msg(pdev->udev, usb_sndctrlpipe(pdev->udev, 0),
		SET_STATUS_CTL,
		USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		LED_FORMATTER,
		pdev->vcinterface,
		&buf, 2, HZ / 2);
}

int pwc_get_leds(struct pwc_device *pdev, int *on_value, int *off_value)
{
	unsigned char buf[2];
	int ret;

	if (pdev->type < 730) {
		*on_value = -1;
		*off_value = -1;
		return 0;
	}

	ret = usb_control_msg(pdev->udev, usb_rcvctrlpipe(pdev->udev, 0),
   	        GET_STATUS_CTL,
		USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,

		LED_FORMATTER,
		pdev->vcinterface,
		&buf, 2, HZ / 2);


	if (ret < 0)
		return ret;
	*on_value = buf[0] * 100;
	*off_value = buf[1] * 100;
	return 0;
}

 /* End of Add-Ons                                    */
 /* ************************************************* */


int pwc_ioctl(struct pwc_device *pdev, unsigned int cmd, void *arg)
{
	switch(cmd) {
	case VIDIOCPWCRUSER:
	{
		if (pwc_restore_user(pdev))
			return -EINVAL;
		break;
	}


	case VIDIOCPWCSUSER:
	{
		if (pwc_save_user(pdev))
			return -EINVAL;
		break;
	}

	case VIDIOCPWCFACTORY:

	{

		if (pwc_restore_factory(pdev))
			return -EINVAL;
		break;
	}

	case VIDIOCPWCSCQUAL:
	{
		int qual, ret;

		if (copy_from_user(&qual, arg, sizeof(int)))
			return -EFAULT;

		if (qual < 0 || qual > 3)
			return -EINVAL;
		ret = pwc_try_video_mode(pdev, pdev->view.x, pdev->view.y, pdev->vframes, qual, pdev->vsnapshot);
		if (ret < 0)
			return ret;

		pdev->vcompression = qual;

		break;

	}
	
	case VIDIOCPWCGCQUAL:
	{
		if (copy_to_user(arg, &pdev->vcompression, sizeof(int)))
			return -EFAULT;
		break;
	}

	case VIDIOCPWCSAGC:
	{
		int agc;
		
		if (copy_from_user(&agc, arg, sizeof(agc)))
			return -EFAULT;
		else {
			if (pwc_set_agc(pdev, agc < 0 ? 1 : 0, agc))
				return -EINVAL;
		}
		break;
	}
	
	case VIDIOCPWCGAGC:
	{
		int agc;
		
		if (pwc_get_agc(pdev, &agc))
			return -EINVAL;
		if (copy_to_user(arg, &agc, sizeof(agc)))
			return -EFAULT;
		break;
	}

	case VIDIOCPWCSSHUTTER:
	{
		int shutter_speed, ret;


		if (copy_from_user(&shutter_speed, arg, sizeof(shutter_speed)))
			return -EFAULT;
		else {
			ret = pwc_set_shutter_speed(pdev, shutter_speed < 0 ? 1 : 0, shutter_speed);
			if (ret < 0)
				return ret;
		}
		break;
	}


 /* ************************************************* */
 /* Begin of Add-Ons for color compensation           */

        case VIDIOCPWCSAWB:
	{
		struct pwc_whitebalance wb;
		int ret;

		if (copy_from_user(&wb, arg, sizeof(wb)))
			return -EFAULT;

		ret = pwc_set_awb(pdev, wb.mode);
		if (ret >= 0 && wb.mode == PWC_WB_MANUAL) {
			pwc_set_red_gain(pdev, wb.manual_red);
			pwc_set_blue_gain(pdev, wb.manual_blue);
		}
		break;
	}

	case VIDIOCPWCGAWB:
	{
		struct pwc_whitebalance wb;
		
		memset(&wb, 0, sizeof(wb));
		wb.mode = pwc_get_awb(pdev);
		if (wb.mode < 0)
			return -EINVAL;
		wb.manual_red = pwc_get_red_gain(pdev);
		wb.manual_blue = pwc_get_blue_gain(pdev);
		if (wb.mode == PWC_WB_AUTO) {
			wb.read_red = pwc_read_red_gain(pdev);
			wb.read_blue = pwc_read_blue_gain(pdev);

		}
		if (copy_to_user(arg, &wb, sizeof(wb)))
			return -EFAULT;
		break;
	}

        case VIDIOCPWCSLED:
	{
		int ret;
		struct pwc_leds leds;

		if (copy_from_user(&leds, arg, sizeof(leds)))
			return -EFAULT;

		ret = pwc_set_leds(pdev, leds.led_on, leds.led_off);
		if (ret<0)
		    return ret;
	    break;
	}



	case VIDIOCPWCGLED:
	{
		int led;
		struct pwc_leds leds;

		led = pwc_get_leds(pdev, &leds.led_on, &leds.led_off);
		if (led < 0)

			return -EINVAL;



		if (copy_to_user(arg, &leds, sizeof(leds)))
			return -EFAULT;
		break;
	}

	//////////////////////////////////////////////////////////////////////

	// Topro IoControl
	//////////////////////////////////////////////////////////////////////
	case VIDIOCWRITEREG:
	{
		unsigned char buff[2];

		if (copy_from_user(buff, arg, 2))
			return -EFAULT;


		topro_write_reg(pdev->udev, buff[0], buff[1]);

		printk("write register index = %.2x, data = %.2x\n", buff[0], buff[1]);

		break;
	}
	case VIDIOCREADREG:
	{
		unsigned char buff[2];

		if (copy_from_user(buff, arg, 2))
			return -EFAULT;

		buff[1] = topro_read_reg(pdev->udev, buff[0]);

		if (copy_to_user(arg, buff, 2))
			return -EFAULT;

		printk("read register index = %.2x, data = %.2x\n", buff[0], buff[1]);
		break;
	}
	case VIDIOCSETGAMMAR:
	{
		if (copy_from_user(tp_gamma_r, arg, GAMMA_TABLE_LENGTH))
			return -EFAULT;

		if (topro_prog_gamma(pdev->udev, 0x00))
		 printk("program gamma R error.\n");

		else

		 printk("program gamma R OK.\n");

		break;

	}
	case VIDIOCSETGAMMAG:
	{
		if (copy_from_user(tp_gamma_g, arg, GAMMA_TABLE_LENGTH))
			return -EFAULT;

		if (topro_prog_gamma(pdev->udev, 0x01))
		 printk("program gamma G error.\n");
		else

		 printk("program gamma G OK.\n");

		break;

	}
	case VIDIOCSETGAMMAB:
	{
		if (copy_from_user(tp_gamma_b, arg, GAMMA_TABLE_LENGTH))
			return -EFAULT;

		if (topro_prog_gamma(pdev->udev, 0x02))
		 printk("program gamma B error.\n");


		else

		 printk("program gamma B OK.\n");

		break;
	}
  
 	case VIDIOCSETEXPOSURE:
	{
    topro_setexposure(pdev->udev,(int)arg);
    printk("SetExposure %d",(int)arg);
		break;
	}
  
 	case VIDIOCSETAE:
  {
    AutoExposure = (int)arg;
    printk("AutoExposure = %d",(int)arg);
    break;
  }   

  case VIDIOCGETSENSORID:
	{
		unsigned char buff[1];

		if (copy_from_user(buff, arg, 1))
			return -EFAULT;
      

#if TP6800_MT9V011
    buff[0] = MT9V011;
#elif TP6800_HV7131R
    buff[0] = HV7131R;
#elif TP6800_PAS302
    buff[0] = PAS302;
#elif TP6800_EVS
    buff[0] = EVS;
#elif TP6800_TAS
    buff[0] = TAS;
#elif TP6800_SHARPCCD
    buff[0] = SHARPCCD;    
#else
   buff[0] = 0;
   
#endif

   printk("SensorID %d\n",buff[0]);
		if (copy_to_user(arg, buff, 1))
			return -EFAULT;
 

		break;
	}

 /* End of Add-Ons                                    */
 /* ************************************************* */

	default:
		return -ENOIOCTLCMD;
		break;
	}
	return 0;
}


////////////////////////////////////////////////////////////////////////////////////
// Topro Programming Subroutines
////////////////////////////////////////////////////////////////////////////////////



void topro_cam_init(struct usb_device *udev)
{
unsigned long i,j;
#ifdef TP6800_QVGA
#if TP6800_EVS // EVS
	// initialize sensor
	topro_setting_sensor(8, 15, 8);	// (Gain,gFrameRate,Exposure)
	// senser
	if (topro_prog_regs(udev, topro_sensor_evs_350a))
		printk("program sensor error.\n");

	// tp68xx initialize data
	if (topro_prog_regs(udev, topro_init_evs_350a))
		printk("program initial data error.\n");
#elif TP6800_TAS // TAS
	// initialize sensor
	if (topro_prog_regs(udev, topro_sensor_tas))
		printk("program sensor error.\n");

	// tp68xx initialize data
	if (topro_prog_regs(udev, topro_init_tas))
		printk("program initial data error.\n");
#elif TP6800_7131E // Hynix 7131E
  // tp68xx initialize data
	if (topro_prog_regs(udev, topro_init_hynix_7131E1))
		printk("program initial data error.\n");

	// VGA parameter
	if (topro_prog_regs(udev, topro_qvga_hynix_7131E1))

		printk("program vga data error.\n");

	// initialize sensor
	if (topro_prog_regs(udev, topro_i2c_hynix_7131E1))

		printk("program sensor error.\n");
#elif TP6800_HV7131R // Hynix 7131R
  // tp68xx initialize data
	if (topro_prog_regs(udev, topro_init_hynix_7131R1))
		printk("program initial data error.\n");

	// VGA parameter
	if (topro_prog_regs(udev, topro_qvga_hynix_7131R1))
		printk("program vga data error.\n");

	// initialize sensor
	if (topro_prog_regs(udev, topro_i2c_hynix_7131R1))

		printk("program sensor error.\n");
#elif TP6800_PAS302 // pixart PAS302
  // tp68xx initialize data
	if (topro_prog_regs(udev, topro_init_pixart_PAS302))
		printk("program initial data error.\n");

	// VGA parameter
	if (topro_prog_regs(udev, topro_qvga_pixart_PAS302))
		printk("program vga data error.\n");

	// initialize sensor

	if (topro_prog_regs(udev, topro_i2c_pixart_PAS302))

		printk("program sensor error.\n");
#elif TP6800_SHARPCCD // pixart PAS302
  // tp68xx initialize data
	if (topro_prog_regs(udev, topro_init_SharpCCD))
		printk("program initial data error.\n");

	// VGA parameter
	if (topro_prog_regs(udev, topro_qvga_SharpCCD))
		printk("program vga data error.\n");

	// initialize sensor
	topro_write_reg(udev, GPIO_IO, 0x91);
	topro_write_reg(udev, GPIO_PU, 0xff);
	topro_write_reg(udev, GPIO_PD, 0x80);

	topro_write_reg(udev, GPIO_DATA, 0x0e);
	topro_write_reg(udev, GPIO_DATA, 0x2e);

    
  
//	sleep(1000);
	topro_write_reg(udev, GPIO_DATA, 0x0e);
//	sleep(100);
	{

		char CCDInit[9]={0x84, 0x01, 0xd9, 0x00, 0x04, 0x1c, 0x01, 0x00, 0x81};
//		CCD_Control(DeviceContext, CCDInit, 72);
	}

	{
		long i, I2C_No;
		unsigned char *pI2C;

		I2C_No = sizeof(topro_init_SharpCCD) / 1;
		pI2C = topro_i2c_SharpCCD;

//		for (i=0;i<I2C_No;i++){
//			CCD_Control(udev, pI2C, 8);

//			pI2C += 1;
//		}
      CCD_Control(udev, pI2C, 8);    
//	  Sleep(500);
	}

    
#else // micron MT9V011
  // tp68xx initialize data
	if (topro_prog_regs(udev, topro_init_micron_MT9V011))
		printk("program initial data error.\n");

  // VGA parameter
	if (topro_prog_regs(udev, topro_qvga_micron_MT9V011))  //930406
		printk("program vga data error.\n");

	// initialize sensor
	if (topro_prog_regs(udev, topro_i2c_micron_MT9V011))
		printk("program sensor error.\n");

#endif
	if (topro_write_reg(udev, FORMAT, 0x01))
		printk("format error.\n");

#else //not TP6800_QVGA
#if TP6800_EVS // EVS
	// initialize sensor
	topro_setting_sensor(8, 15, 8);	// (Gain,gFrameRate,Exposure)
	// senser
	if (topro_prog_regs(udev, topro_sensor_evs_350a))
		printk("program sensor error.\n");

	// tp68xx initialize data
	if (topro_prog_regs(udev, topro_init_evs_350a))
		printk("program initial data error.\n");
#elif TP6800_TAS // TAS
	// initialize sensor
	if (topro_prog_regs(udev, topro_sensor_tas))
		printk("program sensor error.\n");

	// tp68xx initialize data
	if (topro_prog_regs(udev, topro_init_tas))
		printk("program initial data error.\n");
#elif TP6800_7131E // Hynix 7131E

  // tp68xx initialize data
	if (topro_prog_regs(udev, topro_init_hynix_7131E1))
		printk("program initial data error.\n");

	// VGA parameter

	if (topro_prog_regs(udev, topro_vga_hynix_7131E1))
		printk("program vga data error.\n");

	// initialize sensor
	if (topro_prog_regs(udev, topro_i2c_hynix_7131E1))
		printk("program sensor error.\n");
#elif TP6800_HV7131R // Hynix 7131R
  // tp68xx initialize data
	if (topro_prog_regs(udev, topro_init_hynix_7131R1))
		printk("program initial data error.\n");

	// VGA parameter

	if (topro_prog_regs(udev, topro_vga_hynix_7131R1))
		printk("program vga data error.\n");

	// initialize sensor
	if (topro_prog_regs(udev, topro_i2c_hynix_7131R1))
		printk("program sensor error.\n");
#elif TP6800_PAS302 // pixart PAS302
  // tp68xx initialize data
	if (topro_prog_regs(udev, topro_init_pixart_PAS302))
		printk("program initial data error.\n");

	// VGA parameter

	if (topro_prog_regs(udev, topro_vga_pixart_PAS302))
		printk("program vga data error.\n");

	// initialize sensor
	if (topro_prog_regs(udev, topro_i2c_pixart_PAS302))
		printk("program sensor error.\n");
#elif TP6800_SHARPCCD // pixart PAS302
  // tp68xx initialize data
 
	if (topro_prog_regs(udev, topro_init_SharpCCD))
		printk("program initial data error.\n");

	// VGA parameter
	if (topro_prog_regs(udev, topro_vga_SharpCCD))
		printk("program vga data error.\n");

	// initialize sensor
	topro_write_reg(udev, GPIO_IO, 0x91); 
	topro_write_reg(udev, GPIO_PU, 0xff);
	topro_write_reg(udev, GPIO_PD, 0x80);

	topro_write_reg(udev, GPIO_DATA, 0x0e);
//  for (i=0;i<500000000;i++)
  //  j = 555555*555555; 
	topro_write_reg(udev, GPIO_DATA, 0x2e);
  printk("start delay\n");
  for (i=0;i<500000000;i++)   
    j = 555555*555555;  
//	sleep(1000);
  printk("end delay\n");
	topro_write_reg(udev, GPIO_DATA, 0x0e);
  for (i=0;i<50000000;i++)
    j = 555555*555555;
/*
	topro_write_reg(udev, GPIO_DATA, 0x2e);
  printk("start delay\n");
  for (i=0;i<500000000;i++)
    j = 555555*555555;
//	sleep(1000);
  printk("end delay\n");
	topro_write_reg(udev, GPIO_DATA, 0x0e);
  for (i=0;i<500000000;i++)
    j = 555555*555555;
    */
//  sleep(100);
	{
		char CCDInit[9]={0x84, 0x01, 0xd9, 0x00, 0x04, 0x1c, 0x01, 0x00, 0x81};
//		CCD_Control(DeviceContext, CCDInit, 72);
	}

	{
		long i, I2C_No;
		unsigned char *pI2C;


		I2C_No = sizeof(topro_init_SharpCCD) / 1;
		pI2C = topro_i2c_SharpCCD;
		CCD_Control(udev, pI2C, 8);
//    CCD_Control(udev, pI2C, 8);
//		for (i=0;i<I2C_No;i++){
//			CCD_Control(udev, pI2C, 8);
//			pI2C += 1;

//		}
//  sleep(500);
	}
    
#else // micron MT9V011

	if (topro_prog_regs(udev, topro_init_micron_MT9V011))
		printk("program initial data error.\n");
  // VGA parameter

	if (topro_prog_regs(udev, topro_vga_micron_MT9V011))  //930406
		printk("program vga data error.\n");

	// initialize sensor
	if (topro_prog_regs(udev, topro_i2c_micron_MT9V011))
		printk("program sensor error.\n");

  // test image quality
//  topro_write_reg(udev,0x79,0x03);
 
#endif

#endif  //end of TP6800_QVGA

	// Gamma
  
  if (topro_prog_gamma(udev, 0x00))
		printk("program gamma R error.\n");

	if (topro_prog_gamma(udev, 0x01))

		printk("program gamma G error.\n");
	if (topro_prog_gamma(udev, 0x02))
		printk("program gamma B error.\n");


	// Gain
	if (topro_prog_gain(udev, new_gain))
		printk("program gain error.\n");

//	topro_set_hue_saturation(udev, 50, 50);

	// set frame rate
#if TP6800_MT9V011

	if (topro_write_reg(udev, FRAME_RATE, 0x87))
		printk("frame rate error.\n");
#ifdef TP6800_QVGA    
  topro_write_i2c(udev, 0x5d, 0x0a, 0x00, 0x00);  //30fps
#else  
  topro_write_i2c(udev, 0x5d, 0x0a, 0x00, 0x02);  //15fps
#endif
  
#elif TP6800_HV7131R
	if (topro_write_reg(udev, FRAME_RATE, 0x86))
#elif TP6800_SHARPCCD
	if (topro_write_reg(udev, FRAME_RATE, 0x85))  
		printk("frame rate error.\n");
#else  
//	if (topro_write_reg(udev, FRAME_RATE, 0x85))
	if (topro_write_reg(udev, FRAME_RATE, 0x82))  //PAS302
		printk("frame rate error.\n");
#endif    


//Write Default Gamma_R,G,B,Contrast,Brightness,Exposure to File
//  WriteDefaultToFile();        


	printk("Initialize Topro Camera OK!!\n");
}

////////////////////////////////////////////////////////////////////////////////
// Read topro register
////////////////////////////////////////////////////////////////////////////////
unsigned char topro_read_reg(struct usb_device *udev, unsigned char index)
{
	unsigned short buf = 0x0000;
	unsigned char data;
	int ret;

	ret = usb_control_msg(udev, usb_rcvctrlpipe(udev, 0),
		READ_REGISTER,
		READ_REGISTER_TYPE,
		0x0000,
		(buf<< 8) + index,
		&data, 1, HZ);

	return data;
}

////////////////////////////////////////////////////////////////////////////////
// write topro i2c
////////////////////////////////////////////////////////////////////////////////
int topro_write_i2c(struct usb_device *udev, unsigned char dev, unsigned char reg, unsigned char data_h, unsigned char data_l)
{
	topro_write_reg(udev, SIF_ADDR_S, dev);
	topro_write_reg(udev, SIF_ADDR_S2, reg);

	topro_write_reg(udev, SIF_TX_DATA, data_h);
	topro_write_reg(udev, SIF_TX_DATA2, data_l);
	topro_write_reg(udev, SIF_CONTROL, 0x01);

 	return 1;

}
//===================================================================
//      topro_ReadI2C
//===================================================================

int topro_read_i2c(struct usb_device *udev, unsigned char dev, unsigned char reg, unsigned char *data_h, unsigned char *data_l)
{

	topro_write_reg(udev, SIF_ADDR_S, dev);
	topro_write_reg(udev, SIF_ADDR_S2, reg);
	topro_write_reg(udev, SIF_CONTROL, 0x02);  
	*data_h=topro_read_reg(udev, SIF_RX_DATA);
	*data_l=topro_read_reg(udev, SIF_RX_DATA2);


 	return 1;

}


////////////////////////////////////////////////////////////////////////////////
// write topro register
////////////////////////////////////////////////////////////////////////////////
int topro_write_reg(struct usb_device *udev, unsigned char index, unsigned char data)

{

	unsigned short buf = 0x0000;

	return usb_control_msg(udev, usb_sndctrlpipe(udev, 0),
		WRITE_REGISTER,
		WRITE_REGISTER_TYPE,
		(buf<< 8) + data,
		(buf<< 8) + index,
		NULL, 0, HZ);
}

////////////////////////////////////////////////////////////////////////////////
// program topro registers table
////////////////////////////////////////////////////////////////////////////////

int topro_prog_regs(struct usb_device *udev, struct topro_sregs *ptpara)
{

	unsigned short index;

	int ret = 0;


	index = 0;
	do
	{
		if (ptpara[index].index == END_REG) // check end of table
			break;

		// USB control transfer
		ret |= topro_write_reg(udev, ptpara[index].index, ptpara[index].value);

		index++; // next
	} while(index);

	return ret;
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
int topro_setting_iso(struct usb_device *udev, unsigned char enset)
{
	int ret = 0;


	if (enset) // enable

	{
		// open isochronous
		ret |= (topro_write_reg(udev, ENDP_1_CTL, 0x03));
	}
	else // disable
	{
		// close isochronous
		ret |= (topro_write_reg(udev, ENDP_1_CTL, 0x00));
	}

	return ret;
}

////////////////////////////////////////////////////////////////////////////////
// USB bulk out function

////////////////////////////////////////////////////////////////////////////////
#define TOPRO_BULK_LENGTH 31

int topro_bulk_out(struct usb_device *udev, unsigned char bulkctrl, unsigned char *pbulkdata, int bulklen)
{
	int ret = 0;

	unsigned char BulkBuffer[TOPRO_BULK_LENGTH + 1];
	int RetLen;

	int index;
	int fullloop, remainloop;

	//-------------------------------- close isochronous
	topro_setting_iso(udev, 0);


	BulkBuffer[0] = bulkctrl; // set control byte
	fullloop = bulklen / TOPRO_BULK_LENGTH;
	remainloop = bulklen % TOPRO_BULK_LENGTH;

	// full part
	for (index=0; index<fullloop; index++)
	{
	 memcpy(BulkBuffer + 1, pbulkdata + (index * TOPRO_BULK_LENGTH), TOPRO_BULK_LENGTH);
	 ret |= usb_bulk_msg(udev, usb_sndbulkpipe(udev, 3), BulkBuffer, TOPRO_BULK_LENGTH + 1, &RetLen, HZ);

//	 printk("Topro Bulk Status(0) = %d Act = %d \n", ret, RetLen);
	} // for

	// remain part
	if (remainloop)
	{



	 memcpy(BulkBuffer + 1, pbulkdata + (index * TOPRO_BULK_LENGTH), remainloop);
	 ret |= usb_bulk_msg(udev, usb_sndbulkpipe(udev, 3), BulkBuffer, remainloop + 1, &RetLen, HZ);

//	 printk("Topro Bulk Status(1) = %d Act = %d \n", ret, RetLen);

	} // if


	//--------------------------------- open isochronous
	topro_setting_iso(udev, 1);

	return ret;
}

////////////////////////////////////////////////////////////////////////////////
// USB bulk in function
////////////////////////////////////////////////////////////////////////////////

int topro_bulk_in(struct usb_device *udev, unsigned char bulkctrl, unsigned char *pbulkdata, int bulklen)
{
	int ret = 0;

//	unsigned char BulkBuffer[TOPRO_BULK_LENGTH + 1];


  int RetLen;
  

/*
	int index;
	int fullloop, remainloop;

  //-------------------------------- close isochronous
	topro_setting_iso(udev, 0);

	BulkBuffer[0] = bulkctrl; // set control byte
	fullloop = bulklen / TOPRO_BULK_LENGTH;
	remainloop = bulklen % TOPRO_BULK_LENGTH;



	// full part
	for (index=0; index<fullloop; index++)
	{
	 memcpy(BulkBuffer + 1, pbulkdata + (index * TOPRO_BULK_LENGTH), TOPRO_BULK_LENGTH);
	 ret |= usb_bulk_msg(udev, usb_rcvbulkpipe(udev, 2), BulkBuffer, TOPRO_BULK_LENGTH+1, &RetLen, HZ);


//	 printk("Topro Bulk Status(0) = %d Act = %d \n", ret, RetLen);

	} // for

	// remain part
	if (remainloop)
	{
	 memcpy(BulkBuffer + 1, pbulkdata + (index * TOPRO_BULK_LENGTH), remainloop);
	 ret |= usb_bulk_msg(udev, usb_rcvbulkpipe(udev, 2), BulkBuffer, remainloop + 1, &RetLen, HZ);

//	 printk("Topro Bulk Status(1) = %d Act = %d \n", ret, RetLen);
	} // if


	//--------------------------------- open isochronous
	topro_setting_iso(udev, 1);
*/
  
  ret |= usb_bulk_msg(udev, usb_rcvbulkpipe(udev, 2), pbulkdata, bulklen, &RetLen, HZ);
	return ret;
}

////////////////////////////////////////////////////////////////////////////////
// Get CRC value
////////////////////////////////////////////////////////////////////////////////
#define _CRC_G_ 0x8005

unsigned short tp_get_CRC(unsigned char *pbuffer, int bffsize)
{
	int i, j; // loop index

	unsigned char nextbff;
	unsigned short hold, hold_bff;

	hold = 0xffff;

	for (i=0; i<bffsize; i++)
	{
		for (j=0; j<8; j++)
		{
			nextbff = (pbuffer[i] & (0x01 << j)) >> j;
			hold_bff = hold;
			hold = (hold_bff << 1) & 0xfffe;

			if (nextbff != ((hold_bff & 0x8000) >> 15))
				hold = hold ^ _CRC_G_;
		} // for
	} // for

	return ~hold;
}

////////////////////////////////////////////////////
// Calculate CRC Of Gamma Block
////////////////////////////////////////////////////
void tp_cal_CRC(unsigned char bulkctrl, unsigned char *pGammaBlock, int blocklen)

{
	int index;
	unsigned short CRCvalue;
	unsigned char BulkBuffer[TOPRO_BULK_LENGTH + 1];

	BulkBuffer[0] = bulkctrl;
	while(1)
	{
		memcpy(BulkBuffer + 1, pGammaBlock, blocklen);

		CRCvalue = tp_get_CRC(BulkBuffer, blocklen + 1);

	 	if ((((unsigned char)(CRCvalue >> 8) & 0x3F) == 0x3F) ||
			(((unsigned char)CRCvalue & 0x3F) == 0x3F))
		{
			for (index=0; index<blocklen; index++)
				if (pGammaBlock[index] != 0) break;

			if (index == blocklen) // if all items are zeros, infinite loop
				pGammaBlock[index - 1]++;

			else
				pGammaBlock[index]--;

			continue; // next
		} // if
		else
			break;


	} // while
}

////////////////////////////////////////////////////////////////////////////////
// Programming gamma function (set default gamma)
////////////////////////////////////////////////////////////////////////////////
int topro_prog_gamma(struct usb_device *udev, unsigned char bulkctrl)
{
	int index;
	unsigned char *pGamma;

	switch (bulkctrl) // select gamma table
	{
	 case 0x00:
		pGamma = tp_gamma_r;
		break;
	 case 0x01:

		pGamma = tp_gamma_g;
		break;
	 case 0x02:
		pGamma = tp_gamma_b;
		break;
	 default:
		return -1; // fail
	} // switch

	// check ASIC bug
	for (index=0; index<GAMMA_TABLE_LENGTH; index++)
	{
		if (pGamma[index] >= 0xfc)
		{
		 pGamma[index] &= 0xfb;  // ASIC bug, bit 2 must be cleared
		 pGamma[index] |= 0x03;  // ASIC bug
		} // if
	} // for


	// calculate CRC
	int fullloop = GAMMA_TABLE_LENGTH / TOPRO_BULK_LENGTH;
	int remainloop = GAMMA_TABLE_LENGTH % TOPRO_BULK_LENGTH;


	for (index=0; index<fullloop; index++)

	{
		tp_cal_CRC(bulkctrl, pGamma + (index * TOPRO_BULK_LENGTH),
				TOPRO_BULK_LENGTH);
	} // for

	if (remainloop)
	{
		tp_cal_CRC(bulkctrl, pGamma + (index * TOPRO_BULK_LENGTH),
				remainloop);
	} // if

	// bulk out
	return topro_bulk_out(udev, bulkctrl, pGamma, GAMMA_TABLE_LENGTH);
}

//===================================================================
//      WriteGamma
//===================================================================
int topro_write_gamma(struct usb_device *udev,unsigned char *gamma, unsigned char bulkctrl)
{
	int index;
	unsigned char *pGamma;

  pGamma=gamma;

	// check ASIC bug
	for (index=0; index<GAMMA_TABLE_LENGTH; index++)
	{

		if (pGamma[index] >= 0xfc)
		{
		 pGamma[index] &= 0xfb;  // ASIC bug, bit 2 must be cleared

		 pGamma[index] |= 0x03;  // ASIC bug
		} // if
	} // for

	// calculate CRC
	int fullloop = GAMMA_TABLE_LENGTH / TOPRO_BULK_LENGTH;

	int remainloop = GAMMA_TABLE_LENGTH % TOPRO_BULK_LENGTH;

	for (index=0; index<fullloop; index++)

	{
		tp_cal_CRC(bulkctrl, pGamma + (index * TOPRO_BULK_LENGTH),
				TOPRO_BULK_LENGTH);
	} // for

	if (remainloop)
	{
		tp_cal_CRC(bulkctrl, pGamma + (index * TOPRO_BULK_LENGTH),
				remainloop);

	} // if

	// bulk out
	return topro_bulk_out(udev, bulkctrl, pGamma, GAMMA_TABLE_LENGTH);
}



//===================================================================
//      CalculateCBG()
//===================================================================
void CalculateCBG(unsigned char *pData, int Contrast, int Brightness, int Gamma)
{
	int i, j, B, C, G, T, A, Ln_A;
  long L;
	unsigned char DataBuf[256];
	int pstart, pend;
	unsigned char *Src, *Dst;


    B = min(max(Brightness,0),100);
    C = min(max(Contrast,0),100);
    G = min(max(Gamma,10),30);
    G = 1000* 10000 / (G * (50 + B));			//	0.5 < G / 1000 < 4.5
																					//	1/4.5 < 1/(G / 10000) < 1/0.5

    for (i=0;i<256;i++){
		j=198;
		Ln_A = Ln256[i] * G /10000;
		while ((j > 0) && (Ln_A_Table[j] < Ln_A))
			j-=2;
	    T = Ln_A_Table[j+1];			// 0.0521 * 10000 < A_Table < 0.9992 * 10000

		T = (T * (50 + C) * 256) / (1000);
		L = (long)((T + 50) / 100);  // |ˤJ
		*(DataBuf+i) = (unsigned char)max(min(L,255),0);
	}


	Src = DataBuf;
	Dst = pData;
	pstart = 0;
	pend = 1;
	while (pend<255){

		while ((Src[pend] == Src[pstart]) && (pend < 256))
			pend++;
		if (pend < 256){
			Dst[pstart] = Src[pstart];

			if (pend != pstart + 1){
				for (j=1;j<=pend-pstart;j++)
					Dst[pstart+j] = Src[pstart] + (Src[pend] - Src[pstart]) * j / (pend - pstart);
			}
			Dst[pend] = Src[pend];
			pstart = pend;
			pend++;
		}
		else
			for (j=pstart;j<256;j++)
				Dst[j] = Src[j];
	}




}

////////////////////////////////////////////////////////////

//set gamma table for brightness, contrast, gamma
////////////////////////////////////////////////////////////
void topro_set_gamma(struct usb_device *udev)
{
	unsigned char pData[256], Gamma[1024];



//	int YR, YG, YB;
	int i;

  static int framecount=0;
 #define ENDP_1_CTL      0x21
 #define STOP               0
 #define WAKE_UP         0x03  
 #define RED                0
 #define GREEN              1
 #define BLUE               2

 #define AWB_RANGE         50
 static int test=0;
 framecount++;

// printk("topro_framecount %d\n",framecount);
 if (framecount % 30 !=0)
  return;
  
// #define Gamma           0x86
	topro_write_reg(udev,ENDP_1_CTL, STOP);
  if(test==0){
    topro_brightness+=10;
    if (topro_brightness==100)
      test=1;
  }    
  else{
    topro_brightness-=10;
    if (topro_brightness==0)
      test=0;
  }      

	CalculateCBG(pData, topro_contrast, topro_brightness, topro_gamma);
//  printk("topro_Contrast %d\n",topro_contrast);
//  printk("*****topro_Brightness %d\n",topro_brightness);
//  printk("topro_gamma %d\n",topro_gamma);
  

	for (i=0;i<1024;i++)
		Gamma[i] = pData[tp_gamma_r[i]];
		topro_write_gamma(udev,Gamma,RED);


	for (i=0;i<1024;i++)
		Gamma[i] = pData[tp_gamma_g[i]];
		topro_write_gamma(udev,Gamma,GREEN);

	for (i=0;i<1024;i++)
  	Gamma[i] = pData[tp_gamma_b[i]];
		topro_write_gamma(udev,Gamma,BLUE);


	topro_write_reg(udev,ENDP_1_CTL, WAKE_UP);



}



////////////////////////////////////////////////////////////////////////////////
// Programming R/G/B gain function
////////////////////////////////////////////////////////////////////////////////
int topro_prog_gain(struct usb_device *udev, unsigned char *pgain)
{
	return topro_bulk_out(udev, 0x03, pgain, GAIN_TABLE_LENGTH);

}

////////////////////////////////////////////////////////////////////////////////
// matrix3 = matrix1 * matrix2

////////////////////////////////////////////////////////////////////////////////
void matrix_3x3_mul(long matrix1[3][3], long matrix2[3][3], long matrix3[3][3])
{
	int i, j; // loop index

	for (i=0; i<3; i++) {
		for (j=0; j<3; j++) {
			matrix3[i][j] = (matrix1[i][0] * matrix2[0][j]) +
			(matrix1[i][1] * matrix2[1][j]) +

			(matrix1[i][2] * matrix2[2][j]);
		} // for
	} // for
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
void matrix_div_round(long *pmatrix, long count, long unit)
{
	long index; // loop index

	for (index=0; index<count; index++)
	{
		if (*(pmatrix + index) >= 0)
			*(pmatrix + index) = (*(pmatrix + index) + (unit / 2)) / unit;
		else
			*(pmatrix + index) = (*(pmatrix + index) - (unit / 2)) / unit;
	} // for

}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
//void matrix_3x3_print(long *pmatrix)
//{
//	printk("\n Matrix[3][3] = {\n");
//	printk("{%d, %d, %d},\n", (int)pmatrix[0], (int)pmatrix[1], (int)pmatrix[2]);
//	printk("{%d, %d, %d},\n", (int)pmatrix[3], (int)pmatrix[4], (int)pmatrix[5]);
//	printk("{%d, %d, %d}", (int)pmatrix[6], (int)pmatrix[7], (int)pmatrix[8]);
//	printk("};\n");
//}

////////////////////////////////////////////////////////////////////////////////
// Set hue/saturation function
////////////////////////////////////////////////////////////////////////////////
void topro_set_hue_saturation(struct usb_device *udev, long hue, long saturation)
{
	long RGB2YUV[3][3] = { // x 1000
		{ 299,  587,  114},
		{-169, -331,  500},
		{ 500, -418,  -81}};

	long HSB2RGB[3][3] = { // x 100
		{ -98,   38,  341},
		{  24, -183,  231},
		{ -24, -445,  265}};

	long IHSB2RGB[3][3] = { // x 10000
		{-4976,  14827, -6522},
		{ 1091,   1630, -2824},
		{ 1381,   4080, -1560}};


	long XYZ2RGB[3][3] = { // x 10000

		{ 32410, -15374, -4986},
		{ -9692,  18760,   416},
		{   556,  -2040, 10570}};

//	long AWB[3][3] =
//		{  4065,   6183,   -42},
//		{ -1991,  -4631,  5990},
//		{ 20211, -15363, -3523};

	long CCM[3][3] = { // TAS

		{ 5913,    57,  5560},
		{  629,  8752,  2527},
		{-4852, -8344, 26757}};



//	long CCM[3][3] = { // EVS 350
//		{ 4764,  2574,  2164},

//		{ 1695,  7792,   513},
//		{-1346, -1417, 13637}};

	long HUE_SAT_ADJ[3][3] = {
		{0, 0, 0},
		{0, 0, 0},
		{0, 0, 0}};


	long HUE_SAT[3][3];
	long Gain[3][3];

	long matrixbff1[3][3];
	long matrixbff2[3][3];


	int i, j; // loop index
	short gainbulk[9];

	// intialize matrix
	HUE_SAT_ADJ[0][0] = hue * 2;
	HUE_SAT_ADJ[1][1] = saturation * 2;
	HUE_SAT_ADJ[2][2] = 100;


	matrix_3x3_mul(HSB2RGB, HUE_SAT_ADJ, matrixbff1);// x 100 x 10000

	matrix_div_round((long *)matrixbff1, 9, 100); // div 100
//	matrix_3x3_print((long *)matrixbff1);

	matrix_3x3_mul(matrixbff1, IHSB2RGB, HUE_SAT);// 10000, 10000
	matrix_div_round((long *)HUE_SAT, 9, 1000); // div 1000
//	matrix_3x3_print((long *)HUE_SAT);

	matrix_3x3_mul(RGB2YUV, HUE_SAT, matrixbff1); // 1000, 100000
	matrix_div_round((long *)matrixbff1, 9, 1000); // div 1000
//	matrix_3x3_print((long *)matrixbff1);

	matrix_3x3_mul(matrixbff1, XYZ2RGB, matrixbff2); // 100000, 10000
	matrix_div_round((long *)matrixbff2, 9, 10000); // div 10000
//	matrix_3x3_print((long *)matrixbff2);

	matrix_3x3_mul(matrixbff2, CCM, Gain); // 10000, 10000
//	matrix_div_round((long *)Gain, 9, 1000000);
//	matrix_3x3_print((long *)Gain);

//	printk("\nGain Bulk = {\n");
	for (i=0; i<3; i++)
	{
		for (j=0; j<3; j++)
			gainbulk[(i * 3) + j] = (short)maximum(minimum(((Gain[i][j] * 4)
						/ 156250), 511), -512);

//	printk("{%d, %d, %d},\n", gainbulk[(i * 3) + 0], gainbulk[(i * 3) + 1], gainbulk[(i * 3) + 2]);
	} // for
//	printk("};\n");


	topro_prog_gain(udev, (unsigned char *)gainbulk);
}



////////////////////////////////////////////////////////////////////////////////
// calculate sensor parameter
////////////////////////////////////////////////////////////////////////////////
void topro_setting_sensor(int Gain,int gFrameRate,int Exposure)
{
	unsigned short temp1 = 0x0000, temp2 = 0x0000, temp3 = 0x0000, temp4 = 0x0000;
	int i;


	temp3 = 0x0000;
	temp2 = 0x0001;
        temp1 = (unsigned char)(Gain * 2);
	for (i = 0; i<6;i++){
		temp3 <<= 1;
		if (temp2 & temp1)
			temp3 |= 0x04;
		temp2 <<= 1;
	} // for


	switch(gFrameRate){
		case 2:
			temp2 = 1;
			break;
		case 25:
			temp2 = 12;
			break;

		default:

			temp2 = gFrameRate * 2 / 5;
	} // switch

	temp4 = ((Exposure * temp2 * 1000000 / (120 * 2 * 774 / 10)) + 5) / 10 - 1;

	if (temp4 & 0x0001) temp3 += 2;
	if (temp4 & 0x0002) temp3 += 1;
	temp1 = temp4 >> 2;
	temp2 = 0x0001;
	temp4 = 0;
	for (i = 0; i<8;i++) {
		temp4 <<=1;
		if (temp2 & temp1)
			temp4 |= 0x01;
			temp2 <<= 1;
	} // for

	topro_sensor_evs_350a[1].value = temp3; // register index
	topro_sensor_evs_350a[2].value = temp4; // register data
}
///////////////////////////////////////////////////////
//set sensor gain for exposure & AWB
///////////////////////////////////////////////////////
void topro_set_sensorgain(struct usb_device *udev,unsigned char globalgain)
{
  #define MaxGain   0xd0
  #define MinGain   0x05

  
  unsigned char TempGain=0x00, Blue=0x00,Red=0x00;
  unsigned char maxgain,mingain;
  
//  printk("tempgain %d  max %d  min  %d\n",TempGain,maxgain,mingain);
  



  if (AutoWhiteBalance){
      TempGain = (globalgain & 0x7f) * ((globalgain >> 7)+1);
      maxgain  = (MaxGain & 0x7f) * ((MaxGain >> 7)+1);
      mingain  = (MinGain & 0x7f) * ((MinGain >> 7)+1);

      Blue   = TempGain;
      Red    = TempGain;

         
      Blue = Blue * BFactor / 50;
      Red  = Red * RFactor / 50;

      Blue = min(maxgain,max(mingain,Blue));
      Red  = min(maxgain,max(mingain,Red));

      
      if (Blue > 0x7f)
        Blue = ((Blue / 2) | 0x80);
      if (Red > 0x7f)
        Red  = ((Red / 2) | 0x80);


      topro_write_i2c(udev, 0x5d, 0x2b, 0x00, (unsigned char)GlobalGain);
	    topro_write_i2c(udev, 0x5d, 0x2c, 0x00, (unsigned char)Blue);
	    topro_write_i2c(udev, 0x5d, 0x2d, 0x00, (unsigned char)Red);
	    topro_write_i2c(udev, 0x5d, 0x2e, 0x00, (unsigned char)GlobalGain);

      printk("B,R Factor %d %d\n",BFactor,RFactor);
      printk("G1-B-R-G2 %x-%x-%x-%x\n",GlobalGain,Blue,Red,GlobalGain);
      
  }
  else{

	    topro_write_i2c(udev, 0x5d, 0x35, 0x00, (unsigned char)GlobalGain);
      printk("GlobalGain %x\n",GlobalGain);
  }    
  
  
                             
  
  return;
}

void topro_setexposure(struct usb_device *udev, int Exposure)
{
	static int Frequency=60;  //by user selected (indoor ~ 50 or 60 , outdoor ~ 0)
	static int FrameRate=30;    //by user define
	static int ImageSize=VGA;  
  static int SensorVersion=0x8221;

  unsigned char temp1, temp2, temp3;
  long temp4;
  int i;


#if TP6800_MT9V011 
  if (Exposure < 0)
      Exposure = 0;
      
  if (Exposure > 99)
      Exposure = 99;

	unsigned char MT9V011_Exposure[300] = {
				0x01, 0x10, 0xd1, 0x01, 0x14, 0xd1, 0x01, 0x18, 0xd1, 0x01, 0x1c, 0xd1, 0x01, 0x20, 0xd1,
				0x01, 0x24, 0xd1, 0x01, 0x28, 0xd1, 0x01, 0x2c, 0xd1, 0x01, 0x30, 0xd1, 0x01, 0x34, 0xd1,
				0x01, 0x38, 0xd1, 0x01, 0x3c, 0xd1, 0x01, 0xa0, 0xd1, 0x01, 0xa2, 0xd1, 0x01, 0xa4, 0xd1,

				0x01, 0xa6, 0xd1, 0x01, 0xa8, 0xd1, 0x01, 0xaa, 0xd1, 0x01, 0xac, 0xd1, 0x01, 0xae, 0xd1,
				0x01, 0xb0, 0xd1, 0x02, 0x30, 0xd1, 0x02, 0x34, 0xd1, 0x02, 0x38, 0xd1, 0x02, 0x3c, 0xd1,
				0x02, 0xa0, 0xd1, 0x02, 0xa2, 0xd1, 0x02, 0xa4, 0xd1, 0x02, 0xa6, 0xd1, 0x02, 0xa8, 0xd1,

				0x02, 0xaa, 0xd1, 0x02, 0xac, 0xd1, 0x02, 0xae, 0xd1, 0x02, 0xb0, 0xd1, 0x03, 0x3f, 0xd1,

				0x03, 0xa0, 0xd1, 0x03, 0xa2, 0xd1, 0x03, 0xa4, 0xd1, 0x03, 0xa6, 0xd1, 0x03, 0xa8, 0xd1,
				0x03, 0xaa, 0xd1, 0x03, 0xac, 0xd1, 0x03, 0xae, 0xd1, 0x03, 0xb0, 0xd1, 0x04, 0xa4, 0xd1,
				0x04, 0xa6, 0xd1, 0x04, 0xa8, 0xd1, 0x04, 0xaa, 0xd1, 0x04, 0xac, 0xd1, 0x04, 0xae, 0xd1,
				0x04, 0xb0, 0xd1, 0x05, 0xa6, 0xd1, 0x05, 0xa8, 0xd1, 0x05, 0xaa, 0xd1, 0x05, 0xac, 0xd1,
				0x05, 0xae, 0xd1, 0x05, 0xb0, 0xd1, 0x06, 0xa8, 0xd1, 0x06, 0xaa, 0xd1, 0x06, 0xac, 0xd1,

				0x06, 0xae, 0xd1, 0x06, 0xb0, 0xd1, 0x07, 0xaa, 0xd1, 0x07, 0xac, 0xd1, 0x07, 0xae, 0xd1,
				0x07, 0xb0, 0xd1, 0x08, 0xac, 0xd1, 0x08, 0xae, 0xd1, 0x08, 0xb0, 0xd1, 0x09, 0xac, 0xd1,
				0x09, 0xae, 0xd1, 0x09, 0xb0, 0xd1, 0x0a, 0xac, 0xd1, 0x0a, 0xae, 0xd1, 0x0a, 0xb0, 0xd1,
				0x0b, 0xb0, 0xd1, 0x0b, 0xb0, 0xd1, 0x0c, 0xb0, 0xd1, 0x0c, 0xb0, 0xd1, 0x0d, 0xb0, 0xd1,
				0x0d, 0xb0, 0xd1, 0x0e, 0xb0, 0xd1, 0x0e, 0xb0, 0xd1, 0x0f, 0xb0, 0xd1, 0x0f, 0xb0, 0xd1,
				0x10, 0xb0, 0xd1, 0x10, 0xb0, 0xd1, 0x11, 0xb0, 0xd1, 0x11, 0xb0, 0xd1, 0x12, 0xb0, 0xd1,
				0x12, 0xb0, 0xd1, 0x13, 0xb0, 0xd1, 0x13, 0xb0, 0xd1, 0x14, 0xb0, 0xd1, 0x14, 0xb0, 0xd1,

				0x15, 0xb0, 0xd1, 0x16, 0xb0, 0xd1, 0x17, 0xb0, 0xd1, 0x17, 0xb0, 0xd1, 0x17, 0xb0, 0xd1,
	};
  unsigned char MT9V011_Exposure60[300] = {
				0x01, 0x10, 0xd1, 0x01, 0x14, 0xd1, 0x01, 0x18, 0xd1, 0x01, 0x1c, 0xd1, 0x01, 0x20, 0xd1,
				0x01, 0x24, 0xd1, 0x01, 0x28, 0xd1, 0x01, 0x2c, 0xd1, 0x01, 0x30, 0xd1, 0x01, 0x34, 0xd1,
				0x01, 0x38, 0xd1, 0x01, 0x3c, 0xd1, 0x01, 0xa0, 0xd1, 0x01, 0xa2, 0xd1, 0x01, 0xa4, 0xd1,
				0x01, 0xa6, 0xd1, 0x01, 0xa8, 0xd1, 0x01, 0xaa, 0xd1, 0x01, 0xac, 0xd1, 0x01, 0xae, 0xd1,
				0x01, 0xb0, 0xd1, 0x02, 0x30, 0xd1, 0x02, 0x34, 0xd1, 0x02, 0x38, 0xd1, 0x02, 0x3c, 0xd1,
				0x02, 0xa0, 0xd1, 0x02, 0xa2, 0xd1, 0x02, 0xa4, 0xd1, 0x02, 0xa6, 0xd1, 0x02, 0xa8, 0xd1,
				0x02, 0xaa, 0xd1, 0x02, 0xac, 0xd1, 0x02, 0xae, 0xd1, 0x02, 0xb0, 0xd1, 0x03, 0x3f, 0xd1,
				0x03, 0xa0, 0xd1, 0x03, 0xa2, 0xd1, 0x03, 0xa4, 0xd1, 0x03, 0xa6, 0xd1, 0x03, 0xa8, 0xd1,
				0x03, 0xaa, 0xd1, 0x03, 0xac, 0xd1, 0x03, 0xae, 0xd1, 0x03, 0xb0, 0xd1, 0x04, 0xa4, 0xd1,
				0x04, 0xa6, 0xd1, 0x04, 0xa8, 0xd1, 0x04, 0xaa, 0xd1, 0x04, 0xac, 0xd1, 0x04, 0xae, 0xd1,
				0x04, 0xb0, 0xd1, 0x05, 0xa6, 0xd1, 0x05, 0xa8, 0xd1, 0x05, 0xaa, 0xd1, 0x05, 0xac, 0xd1,
				0x05, 0xae, 0xd1, 0x05, 0xb0, 0xd1, 0x06, 0xa8, 0xd1, 0x06, 0xaa, 0xd1, 0x06, 0xac, 0xd1,
				0x06, 0xae, 0xd1, 0x06, 0xb0, 0xd1, 0x07, 0xaa, 0xd1, 0x07, 0xac, 0xd1, 0x07, 0xae, 0xd1,
				0x07, 0xb0, 0xd1, 0x08, 0xac, 0xd1, 0x08, 0xae, 0xd1, 0x08, 0xb0, 0xd1, 0x09, 0xac, 0xd1,
				0x09, 0xae, 0xd1, 0x09, 0xb0, 0xd1, 0x0a, 0xac, 0xd1, 0x0a, 0xae, 0xd1, 0x0a, 0xb0, 0xd1,
				0x0b, 0xb0, 0xd1, 0x0b, 0xb0, 0xd1, 0x0c, 0xb0, 0xd1, 0x0c, 0xb0, 0xd1, 0x0d, 0xb0, 0xd1,
				0x0d, 0xb0, 0xd1, 0x0e, 0xb0, 0xd1, 0x0e, 0xb0, 0xd1, 0x0f, 0xb0, 0xd1, 0x0f, 0xb0, 0xd1,
				0x10, 0xb0, 0xd1, 0x10, 0xb0, 0xd1, 0x11, 0xb0, 0xd1, 0x11, 0xb0, 0xd1, 0x12, 0xb0, 0xd1,
				0x12, 0xb0, 0xd1, 0x13, 0xb0, 0xd1, 0x13, 0xb0, 0xd1, 0x14, 0xb0, 0xd1, 0x14, 0xb0, 0xd1,
				0x15, 0xb0, 0xd1, 0x16, 0xb0, 0xd1, 0x17, 0xb0, 0xd1, 0x17, 0xb0, 0xd1, 0x17, 0xb0, 0xd1,
	};
	unsigned char MT9V011_CIF_Exposure[300] = { 
				0x01, 0x10, 0xd1, 0x01, 0x14, 0xd1, 0x01, 0x18, 0xd1, 0x01, 0x1c, 0xd1, 0x01, 0x20, 0xd1,
				0x01, 0x24, 0xd1, 0x01, 0x28, 0xd1, 0x01, 0x2c, 0xd1, 0x01, 0x30, 0xd1, 0x01, 0x34, 0xd1,
				0x01, 0x38, 0xd1, 0x01, 0x3c, 0xd1, 0x01, 0xa0, 0xd1, 0x01, 0xa2, 0xd1, 0x01, 0xa4, 0xd1,
				0x01, 0xa6, 0xd1, 0x01, 0xa8, 0xd1, 0x01, 0xaa, 0xd1, 0x01, 0xac, 0xd1, 0x01, 0xae, 0xd1,
				0x01, 0xb0, 0xd1, 0x02, 0x30, 0xd1, 0x02, 0x34, 0xd1, 0x02, 0x38, 0xd1, 0x02, 0x3c, 0xd1,
				0x02, 0xa0, 0xd1, 0x02, 0xa2, 0xd1, 0x02, 0xa4, 0xd1, 0x02, 0xa6, 0xd1, 0x02, 0xa8, 0xd1,
				0x02, 0xaa, 0xd1, 0x02, 0xac, 0xd1, 0x02, 0xae, 0xd1, 0x02, 0xb0, 0xd1, 0x03, 0x3f, 0xd1,
				0x03, 0xa0, 0xd1, 0x03, 0xa2, 0xd1, 0x03, 0xa4, 0xd1, 0x03, 0xa6, 0xd1, 0x03, 0xa8, 0xd1,
				0x03, 0xaa, 0xd1, 0x03, 0xac, 0xd1, 0x03, 0xae, 0xd1, 0x03, 0xb0, 0xd1, 0x03, 0xb0, 0xd2, 
				0x03, 0xb0, 0xd3, 0x03, 0xb0, 0xd4, 0x03, 0xb0, 0xd5, 0x03, 0xb0, 0xd6, 0x04, 0xb0, 0xd1,
				0x04, 0xb0, 0xd2, 0x04, 0xb0, 0xd3, 0x04, 0xb0, 0xd4, 0x05, 0xb0, 0xd1, 0x05, 0xb0, 0xd2,
				0x05, 0xb0, 0xd3, 0x05, 0xb0, 0xd4, 0x05, 0xb0, 0xd5, 0x05, 0xb0, 0xd6, 0x05, 0xb0, 0xd7,
				0x06, 0xb0, 0xd4, 0x06, 0xb0, 0xd5, 0x06, 0xb0, 0xd6, 0x07, 0xb0, 0xd3, 0x07, 0xb0, 0xd3,
				0x08, 0xb0, 0xd1, 0x08, 0xb0, 0xd1, 0x08, 0xb0, 0xd2, 0x08, 0xb0, 0xd2, 0x09, 0xae, 0xd1,
				0x09, 0xae, 0xd1, 0x09, 0xb0, 0xd1, 0x0a, 0xac, 0xd1, 0x0a, 0xae, 0xd1, 0x0a, 0xb0, 0xd1,
				0x0b, 0xb0, 0xd1, 0x0b, 0xb0, 0xd1, 0x0c, 0xb0, 0xd1, 0x0c, 0xb0, 0xd1, 0x0d, 0xb0, 0xd1,
				0x0d, 0xb0, 0xd1, 0x0e, 0xb0, 0xd1, 0x0e, 0xb0, 0xd1, 0x0f, 0xb0, 0xd1, 0x0f, 0xb0, 0xd1,
				0x10, 0xb0, 0xd1, 0x10, 0xb0, 0xd1, 0x11, 0xb0, 0xd1, 0x11, 0xb0, 0xd1, 0x12, 0xb0, 0xd1,
				0x12, 0xb0, 0xd1, 0x13, 0xb0, 0xd1, 0x13, 0xb0, 0xd1, 0x14, 0xb0, 0xd1, 0x14, 0xb0, 0xd1,
				0x15, 0xb0, 0xd1, 0x16, 0xb0, 0xd1, 0x17, 0xb0, 0xd1, 0x17, 0xb0, 0xd1, 0x17, 0xb0, 0xd1,
	};
	unsigned char MT9V011_CIF_Exposure60[300] = {
				0x01, 0x10, 0xd1, 0x01, 0x14, 0xd1, 0x01, 0x18, 0xd1, 0x01, 0x1c, 0xd1, 0x01, 0x20, 0xd1,
				0x01, 0x24, 0xd1, 0x01, 0x28, 0xd1, 0x01, 0x2c, 0xd1, 0x01, 0x30, 0xd1, 0x01, 0x34, 0xd1,
				0x01, 0x38, 0xd1, 0x01, 0x3c, 0xd1, 0x01, 0xa0, 0xd1, 0x01, 0xa2, 0xd1, 0x01, 0xa4, 0xd1,
				0x01, 0xa6, 0xd1, 0x01, 0xa8, 0xd1, 0x01, 0xaa, 0xd1, 0x01, 0xac, 0xd1, 0x01, 0xae, 0xd1,
				0x01, 0xb0, 0xd1, 0x02, 0x30, 0xd1, 0x02, 0x34, 0xd1, 0x02, 0x38, 0xd1, 0x02, 0x3c, 0xd1,
				0x02, 0xa0, 0xd1, 0x02, 0xa2, 0xd1, 0x02, 0xa4, 0xd1, 0x02, 0xa6, 0xd1, 0x02, 0xa8, 0xd1,
				0x02, 0xaa, 0xd1, 0x02, 0xac, 0xd1, 0x02, 0xae, 0xd1, 0x02, 0xb0, 0xd1, 0x03, 0x3f, 0xd1,
				0x03, 0xa0, 0xd1, 0x03, 0xa2, 0xd1, 0x03, 0xa4, 0xd1, 0x03, 0xa6, 0xd1, 0x03, 0xa8, 0xd1,
				0x03, 0xaa, 0xd1, 0x03, 0xac, 0xd1, 0x03, 0xae, 0xd1, 0x03, 0xb0, 0xd1, 0x03, 0xb0, 0xd2, 

				0x03, 0xb0, 0xd3, 0x03, 0xb0, 0xd4, 0x03, 0xb0, 0xd5, 0x03, 0xb0, 0xd6, 0x04, 0xb0, 0xd1,
				0x04, 0xb0, 0xd2, 0x04, 0xb0, 0xd3, 0x04, 0xb0, 0xd4, 0x05, 0xb0, 0xd1, 0x05, 0xb0, 0xd2,
				0x05, 0xb0, 0xd3, 0x05, 0xb0, 0xd4, 0x05, 0xb0, 0xd5, 0x05, 0xb0, 0xd6, 0x05, 0xb0, 0xd7,
				0x06, 0xb0, 0xd4, 0x06, 0xb0, 0xd5, 0x06, 0xb0, 0xd6, 0x07, 0xb0, 0xd3, 0x07, 0xb0, 0xd3,
				0x08, 0xb0, 0xd1, 0x08, 0xb0, 0xd1, 0x08, 0xb0, 0xd2, 0x08, 0xb0, 0xd2, 0x09, 0xae, 0xd1,
				0x09, 0xae, 0xd1, 0x09, 0xb0, 0xd1, 0x0a, 0xac, 0xd1, 0x0a, 0xae, 0xd1, 0x0a, 0xb0, 0xd1,

				0x0b, 0xb0, 0xd1, 0x0b, 0xb0, 0xd1, 0x0c, 0xb0, 0xd1, 0x0c, 0xb0, 0xd1, 0x0d, 0xb0, 0xd1,
				0x0d, 0xb0, 0xd1, 0x0e, 0xb0, 0xd1, 0x0e, 0xb0, 0xd1, 0x0f, 0xb0, 0xd1, 0x0f, 0xb0, 0xd1,
				0x10, 0xb0, 0xd1, 0x10, 0xb0, 0xd1, 0x11, 0xb0, 0xd1, 0x11, 0xb0, 0xd1, 0x12, 0xb0, 0xd1,
				0x12, 0xb0, 0xd1, 0x13, 0xb0, 0xd1, 0x13, 0xb0, 0xd1, 0x14, 0xb0, 0xd1, 0x14, 0xb0, 0xd1,
				0x15, 0xb0, 0xd1, 0x16, 0xb0, 0xd1, 0x17, 0xb0, 0xd1, 0x17, 0xb0, 0xd1, 0x17, 0xb0, 0xd1,
	};


	topro_write_i2c(udev, 0x5d, 0x07, 0x00, 0x03);
	topro_write_i2c(udev, 0x5d, 0x0a, 0x00, 0x01);

	if (ImageSize == VGA){

		if (FrameRate == 30)
			i = 2;
		else if (FrameRate == 20)
			i = 3;
		else
			i = 4;
	}
	else
		i = 2;

	topro_write_i2c(udev, 0x5d, 0x07, 0x00, 0x03);
  
	if (Frequency){
		topro_write_i2c(udev, 0x5d, 0x0a, 0x00, i-2);


		if (ImageSize==VGA){
			if (Frequency == 60){
				temp1 = MT9V011_Exposure60[Exposure*3+1];
				temp2 = MT9V011_Exposure60[Exposure*3];
				temp3 = MT9V011_Exposure60[Exposure*3+2];
			}
			else{
				temp1 = MT9V011_Exposure[Exposure*3+1];
				temp2 = MT9V011_Exposure[Exposure*3];
				temp3 = MT9V011_Exposure[Exposure*3+2];
			}
		}
		else{  //QVGA
			if (Frequency == 60){
				temp1 = MT9V011_CIF_Exposure60[Exposure*3+1];
				temp2 = MT9V011_CIF_Exposure60[Exposure*3];
				temp3 = MT9V011_CIF_Exposure60[Exposure*3+2];
			}
			else{
				temp1 = MT9V011_CIF_Exposure[Exposure*3+1];
				temp2 = MT9V011_CIF_Exposure[Exposure*3];
				temp3 = MT9V011_CIF_Exposure[Exposure*3+2];
			}
		}


//		topro_write_i2c(udev, 0x5d, 0x35, 0x00, temp1);
    GlobalGain = temp1;
    topro_set_sensorgain(udev,GlobalGain);
//    printk("Set Global gain %d\n",GlobalGain);
      
		Exposure = temp2;
		if (SensorVersion == 0x8221)
			temp4 = (Exposure * 24000000 / (Frequency * 2) + 224) / (792*i);
		else
			temp4 = (Exposure * 24000000 / (Frequency * 2) + 224) / (786*i);
		
		temp1 = temp4 & 0xff;
		temp2 = temp4 >> 8;
		topro_write_i2c(udev, 0x5d, 0x09, temp2, temp1);
	  if (ImageSize!= VGA)
			topro_write_i2c(udev, 0x5d, 0x41, 0x00, temp3);
      
	}  //Frequency
	else{
		topro_write_i2c(udev, 0x5d, 0x35, 0x00, 0x30);
		temp4 = Exposure;
		temp1 = temp4 & 0xff;
		temp2 = temp4 >> 8;
		topro_write_i2c(udev, 0x5d, 0x09, temp2, temp1);
		topro_write_i2c(udev, 0x5d, 0x41, 0x00, 0xd1);
	}
	topro_write_i2c(udev, 0x5d, 0x07, 0x00, 0x02);
#elif TP6800_SHARPCCD
{
	unsigned char ICX098_CLK[] = {1, 2, 3, 4, 6, 8, 12, 24};
	static int FrameRateOld=0;

	char CCDInit[9]={0x84, 0x01, 0xd9, 0x00, 0x04, 0x1c, 0x01, 0x00, 0x81};
	int CCDExposure;
	unsigned char ICX098_Exposure60[120] = {
		0x01,0xff,0x86, 0x02,0x7a,0x86, 0x02,0xbe,0x86, 0x02,0xff,0x86, 0x03,0xdf,0x86,
		0x03,0xdf,0x86, 0x03,0xff,0x86, 0x04,0xdf,0x86, 0x04,0xdf,0x86, 0x04,0xff,0x86,

		0x05,0xdf,0x85, 0x05,0xdf,0x85, 0x05,0xff,0x85, 0x06,0xff,0x85,

		0x07,0xff,0x84, 0x08,0xff,0x84,

		0x09,0xff,0x83, 0x0a,0xff,0x83, 0x0b,0xff,0x83,	0x0c,0xff,0x83,

		0x0d,0xff,0x82, 0x0e,0xff,0x82, 0x0f,0xff,0x82, 0x10,0xff,0x82,

		0x11,0xff,0x81, 0x12,0xff,0x81, 0x13,0xff,0x81, 0x14,0xff,0x81, 0x15,0xff,0x81,
		0x16,0xff,0x81,	0x17,0xff,0x84, 0x18,0xff,0x81,
		0x1b,0xff,0x80, 0x1e,0xff,0x80, 0x21,0xff,0x80, 0x24,0xff,0x80,
		0x27,0xff,0x80,	0x2a,0xff,0x80, 0x2d,0xff,0x80, 0x30,0xff,0x80,
	};

	unsigned ICX098_Exposure50[120] = {
		0x01,0xff,0x87, 0x02,0x7a,0x87, 0x02,0xbe,0x87, 0x02,0xff,0x87, 0x03,0x7a,0x87,
		0x03,0xbe,0x87, 0x03,0xff,0x87, 0x04,0x7a,0x87, 0x04,0xbe,0x87, 0x04,0xff,0x87,
		0x05,0x7a,0x87, 0x05,0xbe,0x87, 0x05,0xff,0x87,
		0x06,0xff,0x86, 0x07,0xff,0x86,	0x08,0xff,0x86, 0x09,0xff,0x86, 0x0a,0xff,0x86,
		0x0b,0xff,0x85, 0x0c,0xff,0x85,	0x0d,0xff,0x85,
		0x0e,0xff,0x84, 0x0f,0xff,0x84, 0x10,0xff,0x84, 0x11,0xff,0x84,	0x12,0xff,0x84,
		0x13,0xff,0x84, 0x14,0xff,0x84,
		0x15,0xff,0x83, 0x16,0xff,0x83,	0x17,0xff,0x83, 0x18,0xff,0x83, 0x1b,0xff,0x83,
		0x1e,0xff,0x83, 0x21,0xff,0x82,	0x24,0xff,0x82, 0x25,0xff,0x82,
		0x26,0xff,0x82, 0x27,0xff,0x82, 0x28,0xff,0x82,
	};
  Exposure = Exposure * 52 / 99;  
	if (Exposure < 1)
  		Exposure = 1;
	if (Exposure > 52)
	  	Exposure = 52;

	if (Frequency == 50){
		  temp1 = ICX098_Exposure50[(Exposure-1)*3];
		  temp2 = ICX098_Exposure50[(Exposure-1)*3+1];
			temp3 = ICX098_Exposure50[(Exposure-1)*3+2];
			temp4 = ICX098_CLK[temp3 - 0x80];
			CCDExposure = 0x20c - (temp1 * temp4 * 1000000 / 3 / (Frequency * 2) - 447) / 780;
	}
	else if (Frequency == 60){
			temp1 = ICX098_Exposure60[(Exposure-1)*3];
			temp2 = ICX098_Exposure60[(Exposure-1)*3+1];
			temp3 = ICX098_Exposure60[(Exposure-1)*3+2];
	}
	else{
			temp2 = 0x80;
			temp3 = 0x87;
			CCDExposure = Exposure << 1;
	}

//				temp4 = temp2;
				if (Exposure <=12)
					temp3 = 0x86;
				else if (Exposure <=18)
					temp3 = 0x85;
				else if (Exposure <=24)
					temp3 = 0x84;
				else if (Exposure <=28)
					temp3 = 0x83;
				else if (Exposure <=32)
					temp3 = 0x82;
				else
					temp3 = 0x81;

				topro_write_reg(udev, FRAME_RATE, temp3);

				CCDInit[0] = (CCDExposure << 2) | 0x01;
				CCDInit[1] = (CCDExposure >> 6) & 0x0f;

				CCDInit[0] = Exposure;
				if (FrameRateOld != temp3){
					FrameRateOld = temp3;
					CCDInit[0] |= 0x80;
				}
				CCD_Control(udev, CCDInit, 8);
}
#elif TP6800_HV7131R
//	topro_write_i2c(udev, 0x11, 0x20, 0x02, 0x00);
//	topro_write_i2c(udev, 0x11, 0x21, 0x46, 0x00);
//	topro_write_i2c(udev, 0x11, 0x25, 0x09, 0x00);
//	topro_write_i2c(udev, 0x11, 0x26, 0x27, 0x00);
//	topro_write_i2c(udev, 0x11, 0x27, 0xc0, 0x00);

	unsigned char HV7131R1_Exposure60[200] = {
				0x01, 0x02, 0x01, 0x02, 0x01, 0x04, 0x01, 0x08, 0x01, 0x0b,  //5
				0x01, 0x0d, 0x02, 0x03, 0x02, 0x04, 0x02, 0x05, 0x02, 0x06,  //10
				0x02, 0x07, 0x02, 0x08, 0x03, 0x03, 0x03, 0x04, 0x03, 0x04,  //15
				0x03, 0x05, 0x03, 0x06, 0x03, 0x07, 0x03, 0x08, 0x04, 0x05,  //20
				0x04, 0x06, 0x04, 0x07, 0x04, 0x08, 0x04, 0x09, 0x04, 0x0a,  //25
				0x04, 0x0b, 0x04, 0x0c, 0x05, 0x09, 0x05, 0x0a, 0x05, 0x0b,  //30
				0x05, 0x0c, 0x05, 0x0d, 0x05, 0x0e, 0x05, 0x0f, 0x05, 0x10,  //35
				0x06, 0x0d, 0x06, 0x0e, 0x06, 0x0f, 0x06, 0x10, 0x06, 0x11,  //40
				0x06, 0x12, 0x06, 0x13, 0x06, 0x14, 0x07, 0x11, 0x07, 0x12,  //45
				0x07, 0x13, 0x07, 0x14, 0x07, 0x15, 0x08, 0x12, 0x08, 0x13,  //50
				0x08, 0x14, 0x08, 0x15, 0x08, 0x16, 0x08, 0x17, 0x08, 0x18,  //55
				0x08, 0x19, 0x09, 0x16, 0x09, 0x17, 0x09, 0x18, 0x09, 0x19,  //60
				0x09, 0x1a, 0x0a, 0x17, 0x0a, 0x18, 0x0a, 0x19, 0x0a, 0x1a,  //65
				0x0a, 0x1b, 0x0b, 0x18, 0x0b, 0x19, 0x0b, 0x1a, 0x0b, 0x1b,  //70
				0x0b, 0x1c, 0x0c, 0x1a, 0x0c, 0x1b, 0x0c, 0x1c, 0x0c, 0x1d,  //75
				0x0c, 0x1e, 0x0d, 0x1c, 0x0d, 0x1d, 0x0e, 0x1b, 0x0e, 0x1c,  //80
				0x0f, 0x1a, 0x0f, 0x1b, 0x10, 0x19, 0x10, 0x1a, 0x11, 0x19,  //85

				0x11, 0x1a, 0x12, 0x19, 0x12, 0x1a, 0x13, 0x19, 0x13, 0x1a,  //90
				0x14, 0x19, 0x14, 0x1a, 0x15, 0x19, 0x15, 0x1a, 0x16, 0x19,  //95
				0x16, 0x1a, 0x17, 0x19, 0x17, 0x1a, 0x18, 0x19, 0x18, 0x1a,  //100
	};
  
  if (Exposure < 0)
      Exposure = 0;

  if (Exposure > 99)
      Exposure = 99;
	if (Frequency){
  		temp4 = HV7131R1_Exposure60[Exposure*2] * ((12000000 * (FrameRate == 20 ? 30 : FrameRate) / 30) / (Frequency * 2));
  }
	else
	    temp4 = Exposure * 8 * 256;
      
  temp1 = temp4 & 0xff;
	temp2 = (temp4 >> 8) & 0xff;
	temp3 = (temp4 >> 16) & 0xff;

	if (FrameRate == 20){
			topro_write_i2c(udev, 0x11, 0x20, 0x02, 0x00);
			topro_write_i2c(udev, 0x11, 0x21, 0x46, 0x00);
	}
	else{
			topro_write_i2c(udev, 0x11, 0x21, 0xd0, 0x00);
			topro_write_i2c(udev, 0x11, 0x20, 0x00, 0x00);
	}      
      

	topro_write_i2c(udev, 0x11, 0x27, temp1, 0x00);
	topro_write_i2c(udev, 0x11, 0x26, temp2, 0x00);
	topro_write_i2c(udev, 0x11, 0x25, temp3, 0x00);

	topro_write_i2c(udev, 0x11, 0x30, (unsigned char)HV7131R1_Exposure60[Exposure*2+1], 0x00);


  return;
#else  // TP6800_PAS302

	unsigned char PAS302_CIF_Exposure60[256] = {
		//		Np,	ExpH, ExpL, Gain
				0x02, 0xB7, 0x00, 0x00, 0x02, 0xB7, 0x00, 0x01, 0x02, 0xB7, 0x00, 0x02, 0x02, 0xB7, 0x00, 0x03, 0x02, 0xB7, 0x00, 0x04,
				0x02, 0xB7, 0x00, 0x05, 0x02, 0xB7, 0x00, 0x06,	0x02, 0xB7, 0x00, 0x07,	0x02, 0xB7, 0x00, 0x08,	0x02, 0xB7, 0x00, 0x09,
				0x02, 0xB7, 0x00, 0x0a, 0x02, 0xB7, 0x00, 0x0b,	0x02, 0xB7, 0x00, 0x0c,	0x02, 0xB7, 0x00, 0x0c,	0x02, 0xB7, 0x00, 0x0c,
				0x02, 0x7A, 0x00, 0x01, 0x02, 0x7A, 0x00, 0x01, 0x02, 0x7A, 0x00, 0x02, 0x02, 0x7A, 0x00, 0x03, 0x02, 0x7A, 0x00, 0x04,
				0x02, 0x7A, 0x00, 0x05, 0x02, 0x7A, 0x00, 0x06, 0x02, 0x7A, 0x00, 0x07, 0x02, 0x3D, 0x00, 0x01, 0x02, 0x3D, 0x00, 0x01,
				0x02, 0x3D, 0x00, 0x02, 0x02, 0x3D, 0x00, 0x03, 0x02, 0x3D, 0x00, 0x04, 0x02, 0x3D, 0x00, 0x05, 0x03, 0x51, 0x00, 0x00,


				0x03, 0x51, 0x00, 0x01, 0x03, 0x51, 0x00, 0x02, 0x03, 0x51, 0x00, 0x03, 0x03, 0x28, 0x01, 0x00,	0x03, 0x28, 0x01, 0x01,
				0x03, 0x28, 0x01, 0x02, 0x04, 0x3D, 0x00, 0x00, 0x04, 0x3D, 0x00, 0x01, 0x04, 0x1E, 0x01, 0x00,	0x04, 0x1E, 0x01, 0x01,
				0x06, 0x51, 0x00, 0x00,	0x06, 0x51, 0x00, 0x01, 0x06, 0x3D, 0x00, 0x00,	0x06, 0x28, 0x01, 0x00,	0x06, 0x14, 0x00, 0x00,
				0x08, 0x3D, 0x00, 0x00,	0x08, 0x2D, 0x01, 0x00, 0x08, 0x1E, 0x01, 0x00,	0x08, 0x0F, 0x00, 0x00,	0x0C, 0x51, 0x00, 0x00,
				0x0C, 0x47, 0x00, 0x00,	0x0C, 0x3D, 0x00, 0x00, 0x0C, 0x32, 0x01, 0x00,	0x0C, 0x28, 0x01, 0x00,	0x0C, 0x1E, 0x01, 0x00,
				0x0C, 0x14, 0x00, 0x00,	0x0C, 0x0A, 0x00, 0x00, 0x0C, 0x0A, 0x00, 0x01, 0x0C, 0x0A, 0x00, 0x02,	0x0C, 0x0A, 0x00, 0x03,
				0x0C, 0x0A, 0x00, 0x04,	0x0C, 0x0A, 0x00, 0x07, 0x0C, 0x0A, 0x00, 0x09, 0x0C, 0x0A, 0x00, 0x0a
//0 ~ 29 /30f
//30 ~ 35 /20f
//36 ~ 39 /15f
//40 ~ 44 /10f
//45 ~ 48 /7.5f
//49 ~ 56 /5f

	};
	unsigned char PAS302_VGA_Exposure60[256] = {
		//		|						|						|						|						|
				0x03, 0xcb, 0x01, 0x00, 0x03, 0xcb, 0x01, 0x01, 0x03, 0xcb, 0x01, 0x02, 0x03, 0xcb, 0x01, 0x03, 0x03, 0xcb, 0x01, 0x04,
				0x03, 0xcb, 0x01, 0x05, 0x03, 0xcb, 0x01, 0x06,	0x03, 0xcb, 0x01, 0x07,	0x03, 0xcb, 0x01, 0x08,	0x03, 0xcb, 0x01, 0x09,
				0x03, 0xcb, 0x01, 0x0a, 0x03, 0xcb, 0x01, 0x0b,	0x03, 0xcb, 0x01, 0x0c,	0x03, 0xcb, 0x01, 0x0c,	0x03, 0xcb, 0x01, 0x0c,
				0x03, 0xa2, 0x01, 0x01, 0x03, 0xa2, 0x01, 0x01, 0x03, 0xa2, 0x01, 0x02, 0x03, 0xa2, 0x01, 0x03, 0x03, 0xa2, 0x01, 0x04,

				0x03, 0xa2, 0x01, 0x05, 0x03, 0xa2, 0x01, 0x06, 0x03, 0xa2, 0x01, 0x07, 0x03, 0x7a, 0x00, 0x01, 0x03, 0x7a, 0x00, 0x01,
				0x03, 0x7a, 0x00, 0x02, 0x03, 0x7a, 0x00, 0x03, 0x03, 0x7a, 0x00, 0x04, 0x03, 0x7a, 0x00, 0x05, 0x03, 0x51, 0x00, 0x00,
				0x03, 0x51, 0x00, 0x01, 0x03, 0x51, 0x00, 0x02, 0x03, 0x51, 0x00, 0x03, 0x03, 0x28, 0x01, 0x00,	0x03, 0x28, 0x01, 0x01,
				0x03, 0x28, 0x01, 0x02, 0x04, 0x3D, 0x00, 0x00, 0x04, 0x3D, 0x00, 0x01, 0x04, 0x1E, 0x01, 0x00,	0x04, 0x1E, 0x01, 0x01,
				0x06, 0x51, 0x00, 0x00,	0x06, 0x51, 0x00, 0x01, 0x06, 0x3D, 0x00, 0x00,	0x06, 0x28, 0x01, 0x00,	0x06, 0x14, 0x00, 0x00,
				0x08, 0x3D, 0x00, 0x00,	0x08, 0x2D, 0x01, 0x00, 0x08, 0x1E, 0x01, 0x00,	0x08, 0x0F, 0x00, 0x00,	0x0C, 0x51, 0x00, 0x00,
				0x0C, 0x47, 0x00, 0x00,	0x0C, 0x3D, 0x00, 0x00, 0x0C, 0x32, 0x01, 0x00,	0x0C, 0x28, 0x01, 0x00,	0x0C, 0x1E, 0x01, 0x00,
				0x0C, 0x14, 0x00, 0x00,	0x0C, 0x0A, 0x00, 0x00, 0x0C, 0x0A, 0x00, 0x01, 0x0C, 0x0A, 0x00, 0x02,	0x0C, 0x0A, 0x00, 0x03,
				0x0C, 0x0A, 0x00, 0x04,	0x0C, 0x0A, 0x00, 0x07, 0x0C, 0x0A, 0x00, 0x09, 0x0C, 0x0A, 0x00, 0x0a


//0 ~ 35 /20f
//36 ~ 39 /15f
//40 ~ 44 /10f
//45 ~ 48 /7.5f
//49 ~ 56 /5f
	};
	unsigned char PAS302_CIF_Exposure50[256] = {
		//		|						|						|						|						|
				0x02, 0xAA, 0x01, 0x00, 0x02, 0xAA, 0x01, 0x01, 0x02, 0xAA, 0x01, 0x02, 0x02, 0xAA, 0x01, 0x03, 0x02, 0xAA, 0x01, 0x04,
				0x02, 0xAA, 0x01, 0x05, 0x02, 0xAA, 0x01, 0x06, 0x02, 0xAA, 0x01, 0x07, 0x02, 0xAA, 0x01, 0x08, 0x02, 0xAA, 0x01, 0x09,
				0x02, 0xAA, 0x01, 0x0a, 0x02, 0xAA, 0x01, 0x0b, 0x02, 0xAA, 0x01, 0x0c, 0x02, 0xAA, 0x01, 0x0c, 0x02, 0xAA, 0x01, 0x0c,
				0x02, 0x61, 0x01, 0x01, 0x02, 0x61, 0x01, 0x01, 0x02, 0x61, 0x01, 0x02, 0x02, 0x61, 0x01, 0x03, 0x02, 0x61, 0x01, 0x04,
				0x02, 0x61, 0x01, 0x05, 0x02, 0x61, 0x01, 0x06, 0x02, 0x61, 0x01, 0x07, 0x02, 0x18, 0x00, 0x01, 0x02, 0x18, 0x00, 0x01,
				0x02, 0x18, 0x00, 0x02, 0x02, 0x18, 0x00, 0x03, 0x02, 0x18, 0x00, 0x04, 0x02, 0x18, 0x00, 0x05, 0x03, 0x30, 0x01, 0x00,

				0x03, 0x30, 0x01, 0x01, 0x03, 0x30, 0x01, 0x02, 0x03, 0x30, 0x01, 0x03, 0x04, 0x3D, 0x00, 0x00,	0x04, 0x3D, 0x00, 0x01,
				0x04, 0x3D, 0x00, 0x02, 0x04, 0x18, 0x00, 0x00, 0x04, 0x18, 0x00, 0x01, 0x06, 0x49, 0x00, 0x00, 0x06, 0x49, 0x00, 0x01,
				0x06, 0x30, 0x01, 0x00, 0x06, 0x30, 0x01, 0x01, 0x06, 0x18, 0x00, 0x00, 0x08, 0x3D, 0x00, 0x00,	0x08, 0x2A, 0x01, 0x00,
				0x08, 0x18, 0x00, 0x00,	0x08, 0x06, 0x00, 0x00, 0x0C, 0x49, 0x00, 0x00, 0x0C, 0x3D, 0x00, 0x00,	0x0C, 0x30, 0x01, 0x00,
				0x0C, 0x24, 0x01, 0x00,	0x0C, 0x18, 0x00, 0x00, 0x0C, 0x0C, 0x00, 0x00, 0x0C, 0x0C, 0x00, 0x01,	0x0C, 0x0C, 0x00, 0x02,
				0x0C, 0x0C, 0x00, 0x03,	0x0C, 0x0C, 0x00, 0x04, 0x0C, 0x0C, 0x00, 0x05, 0x0C, 0x0C, 0x00, 0x06, 0x0C, 0x0C, 0x00, 0x07,
				0x0C, 0x0C, 0x00, 0x08,	0x0C, 0x0C, 0x00, 0x09, 0x0C, 0x0C, 0x00, 0x0a, 0x0C, 0x0C, 0x00, 0x0b
	};
	unsigned char PAS302_VGA_Exposure50[256] = {
		//		|						|						|						|						|
				0x03, 0xc3, 0x00, 0x00, 0x03, 0xc3, 0x00, 0x01, 0x03, 0xc3, 0x00, 0x02, 0x03, 0xc3, 0x00, 0x03, 0x03, 0xc3, 0x00, 0x04,
				0x03, 0xc3, 0x00, 0x05, 0x03, 0xc3, 0x00, 0x06, 0x03, 0xc3, 0x00, 0x07, 0x03, 0xc3, 0x00, 0x08, 0x03, 0xc3, 0x00, 0x09,
				0x03, 0xc3, 0x00, 0x0a, 0x03, 0xc3, 0x00, 0x0b, 0x03, 0xc3, 0x00, 0x0c, 0x03, 0xc3, 0x00, 0x0c, 0x03, 0xc3, 0x00, 0x0c,
				0x03, 0x92, 0x01, 0x01, 0x03, 0x92, 0x01, 0x01, 0x03, 0x92, 0x01, 0x02, 0x03, 0x92, 0x01, 0x03, 0x03, 0x92, 0x01, 0x04,
				0x03, 0x92, 0x01, 0x05, 0x03, 0x92, 0x01, 0x06, 0x03, 0x92, 0x01, 0x07, 0x03, 0x61, 0x00, 0x01, 0x03, 0x61, 0x00, 0x01,

				0x03, 0x61, 0x00, 0x02, 0x03, 0x61, 0x00, 0x03, 0x03, 0x61, 0x00, 0x04, 0x03, 0x61, 0x00, 0x05, 0x03, 0x30, 0x01, 0x00,

				0x03, 0x30, 0x01, 0x01, 0x03, 0x30, 0x01, 0x02, 0x03, 0x30, 0x01, 0x03, 0x04, 0x3D, 0x00, 0x00,	0x04, 0x3D, 0x00, 0x01,
				0x04, 0x3D, 0x00, 0x02, 0x04, 0x18, 0x00, 0x00, 0x04, 0x18, 0x00, 0x01, 0x06, 0x49, 0x00, 0x00, 0x06, 0x49, 0x00, 0x01,
				0x06, 0x30, 0x01, 0x00, 0x06, 0x30, 0x01, 0x01, 0x06, 0x18, 0x00, 0x00, 0x08, 0x3D, 0x00, 0x00,	0x08, 0x2A, 0x01, 0x00,
				0x08, 0x18, 0x00, 0x00,	0x08, 0x06, 0x00, 0x00, 0x0C, 0x49, 0x00, 0x00, 0x0C, 0x3D, 0x00, 0x00,	0x0C, 0x30, 0x01, 0x00,
				0x0C, 0x24, 0x01, 0x00,	0x0C, 0x18, 0x00, 0x00, 0x0C, 0x0C, 0x00, 0x00, 0x0C, 0x0C, 0x00, 0x01,	0x0C, 0x0C, 0x00, 0x02,
				0x0C, 0x0C, 0x00, 0x03,	0x0C, 0x0C, 0x00, 0x04, 0x0C, 0x0C, 0x00, 0x05, 0x0C, 0x0C, 0x00, 0x06, 0x0C, 0x0C, 0x00, 0x07,
				0x0C, 0x0C, 0x00, 0x08,	0x0C, 0x0C, 0x00, 0x09, 0x0C, 0x0C, 0x00, 0x0a, 0x0C, 0x0C, 0x00, 0x0b
	};
//  int exposure;
  Exposure = Exposure * 64 / 99;
  if (Exposure < 0)
      Exposure = 0;

  if (Exposure > 63)
      Exposure = 63;
      
				unsigned char RegData=0x00;
				unsigned char tempGain;

					if (!Frequency){
						int offset_ny;
						if (FrameRate == 25)
							offset_ny = 497 - Exposure;
						else
							offset_ny = 486 - Exposure;
						temp1= ((offset_ny/2) & 0xFF);
						temp2 = (offset_ny & 0x01);
						temp3 = 0x4c;	//this is not important
					  topro_write_i2c(udev, 0x40, 0x0F, temp1, 0x00);
					//offset_ny bit0
					  topro_write_i2c(udev, 0x40, 0x0E, temp2, 0x00);
					//offset_ne_by4 bit7-0
					  topro_write_i2c(udev, 0x40, 0x0D, temp3, 0x00);

					  topro_write_i2c(udev, 0x40, 0x11, 0x01, 0x00);
					}
					else{
						if (ImageSize != VGA){
							if (Frequency == 60){
								temp1 = PAS302_CIF_Exposure60[Exposure*4];
								temp2 = PAS302_CIF_Exposure60[Exposure*4+1];
								temp3 = PAS302_CIF_Exposure60[Exposure*4+2];
								tempGain = PAS302_CIF_Exposure60[Exposure*4+3];
							}
							else{
								temp1 = PAS302_CIF_Exposure50[Exposure*4];

								temp2 = PAS302_CIF_Exposure50[Exposure*4+1];
								temp3 = PAS302_CIF_Exposure50[Exposure*4+2];
								tempGain = PAS302_CIF_Exposure50[Exposure*4+3];
							}
	//ethan do it
						}
						else{
							if (Frequency == 60){
								temp1 = PAS302_VGA_Exposure60[Exposure*4];
								temp2 = PAS302_VGA_Exposure60[Exposure*4+1];
								temp3 = PAS302_VGA_Exposure60[Exposure*4+2];
								tempGain = PAS302_VGA_Exposure60[Exposure*4+3];
							}
							else{

								temp1 = PAS302_VGA_Exposure50[Exposure*4];
								temp2 = PAS302_VGA_Exposure50[Exposure*4+1];
								temp3 = PAS302_VGA_Exposure50[Exposure*4+2];
								tempGain = PAS302_VGA_Exposure50[Exposure*4+3];

							}

						}
/*
						if (ImageSize == VGA){
							if (DC->CPUSpeed < 1000){
								if (temp1 < 0x06) //10 fps
									temp1 = 0x06;
							}
							else if (DC->CPUSpeed < 1800){
								if (temp1 < 0x04){ //15fps
									temp1 = 0x04;
								}

//						sprintf(szDbg, "CPU %d temp1 %x", DC->CPUSpeed, temp1);
//						WriteImageInfoToFile(DC, szDbg);
							}

						}
						else{
							if (DC->CPUSpeed < 1000){
								if (temp1 < 0x04) //15 fps
									temp1 = 0x04;
							}
							else if (DC->CPUSpeed < 1800){
								if (temp1 < 0x03) //20fps
									temp1 = 0x03;
							}
						}
*/

//						sprintf(szDbg, "SetExposure %d %x %x %x %x", Exposure, temp1, temp2, temp3,tempGain);
//						WriteImageInfoToFile(DC, szDbg);

//						ReadRegister(DC,0x11, &RegData);

						topro_write_i2c(udev, 0x40, 0x11, 0x00, 0x00);


						topro_write_i2c(udev, 0x40, 0x02, temp1, 0x00);
						//offset_ny bit0
						topro_write_i2c(udev, 0x40, 0x0f, temp2, 0x00);
						//offset_ne_by4 bit7-0
						topro_write_i2c(udev, 0x40, 0x0e, temp3, 0x00);
						topro_write_i2c(udev, 0x40, 0x10, tempGain, 0x00);

						topro_write_i2c(udev, 0x40, 0x11, 0x01, 0x00);
					}


				
      

   
#endif 




}


int CurrentExposure=10;
int PreExposure = 0;

int FrameSkipCount=0;
//
void topro_autoexposure(struct usb_device *udev)
{
  
// int ntstatus=0;

 long AccY;
 static int FrameSkipCount=0;

 unsigned char BulkInData[32];
 static unsigned char *pBulkInData;
 pBulkInData = BulkInData;
 int RateAE = 6;
// static int test=0;
// unsigned char TempY1,TempY2,TempY3,TempY4;
#define AEThreshold 120

//  printk("AutoExposure\n");
  FrameSkipCount++;
  
  if (AutoExposure==0)
      return;
  
  if (FrameSkipCount % RateAE !=0)
    return;



  topro_bulk_in(udev, 0x00, pBulkInData, 32);
  printk("topro_bulk_in %x %x %x %x\n",BulkInData[0],BulkInData[1],BulkInData[2],BulkInData[3]);
 

  PreExposure = CurrentExposure;



  AccY  = (BulkInData[19] << 24) | (BulkInData[18] << 16)
               | (BulkInData[17] << 8) | BulkInData[16];
               
  AccY = AccY / (640 * 480);
  
  if (AccY > 255 || AccY <= 0){
    printk("AccY bulk_in data Error\n");
    return;
  }

  
 printk("AccY %d\n",(int)AccY);
  if (AccY > AEThreshold + 2){
    if (AccY > AEThreshold + 20){
       CurrentExposure -= 5;

    }
    else
      CurrentExposure -= 1;


  }
  else if ( AccY < AEThreshold - 2){
    if (AccY < AEThreshold - 20){
       CurrentExposure += 5;

    }
    else
        CurrentExposure += 1;
  }

  printk("Exposure %d\n ",CurrentExposure);

/* 
  if(test==0){
    CurrentExposure+=10;
    if (CurrentExposure==100)
        test=1;
  }
  else{

    CurrentExposure-=10;
    if (CurrentExposure==0)
        test=0;
  }        
*/
  printk("CurrentExposure %d\n",CurrentExposure);
//  if (CurrentExposure != PreExposure){
    topro_setexposure(udev,CurrentExposure);
//  }

}
/*
////////////////////////////////////////////////////////////////////////////////
// Set AWB function  (by modify gamma table)
////////////////////////////////////////////////////////////////////////////////
int topro_set_awb(struct usb_device *udev, unsigned char bulkctrl, int factor)
{


	int index,tempGamma;
	unsigned char pGamma[1024];

	switch (bulkctrl) // select gamma table
	{
	 case 0x00:
      memcpy(&pGamma,tp_gamma_r,1024);   
		break;
	 case 0x01:
      memcpy(&pGamma,tp_gamma_g,1024); 
		break;
	 case 0x02:

      memcpy(&pGamma,tp_gamma_b,1024);  
		break;
	 default:
		return -1; // fail
	} // switch

	// check ASIC bug
	for (index=0; index<GAMMA_TABLE_LENGTH; index++)
	{
    tempGamma= pGamma[index];
    if ((tempGamma * (factor + 50) / 50) > 255)
      pGamma[index] = 255;
    else  
      pGamma[index] = tempGamma * (factor + 50) / 50;
            
		if (pGamma[index] >= 0xfc)
		{
		 pGamma[index] &= 0xfb;  // ASIC bug, bit 2 must be cleared
		 pGamma[index] |= 0x03;  // ASIC bug
		} // if
	} // for

	// calculate CRC
	int fullloop = GAMMA_TABLE_LENGTH / TOPRO_BULK_LENGTH;
	int remainloop = GAMMA_TABLE_LENGTH % TOPRO_BULK_LENGTH;

	for (index=0; index<fullloop; index++)
	{
		tp_cal_CRC(bulkctrl, pGamma + (index * TOPRO_BULK_LENGTH),
				TOPRO_BULK_LENGTH);
	} // for

	if (remainloop)

	{
		tp_cal_CRC(bulkctrl, pGamma + (index * TOPRO_BULK_LENGTH),
				remainloop);
	} // if


	// bulk out
	return topro_bulk_out(udev, bulkctrl, pGamma, GAMMA_TABLE_LENGTH);
}
*/

/////////////////////////////////////
//AutoWhiteBalance Function
/////////////////////////////////////
void topro_autowhitebalance(struct usb_device *udev)
{
 if (AutoWhiteBalance==0)
    return;
 
 long AccR=0,AccG=0,AccB=0;
 unsigned char BulkInData[32];
 static unsigned char *pBulkInData;
 pBulkInData = BulkInData;
// unsigned char i2c1,i2c2;
 int RateCount = 6;
 //unsigned char *pgain,*data1,*data2;
// static int test=0;
  FrameSkipCount++;

  if (FrameSkipCount % RateCount !=0)
    return;


//  topro_read_i2c(udev,0x5d,0x2c,&i2c1,&i2c2);
//  printk("i2c1 %x  i2c2 %x\n",i2c1,i2c2);

      
  topro_bulk_in(udev, 0x00, pBulkInData, 32); 
    
  AccR  = (BulkInData[3] << 24) | (BulkInData[2] << 16)

               | (BulkInData[1] << 8) | BulkInData[0];

  AccG  = (BulkInData[7] << 24) | (BulkInData[6] << 16)
               | (BulkInData[5] << 8) | BulkInData[4];

  AccB  = (BulkInData[11] << 24) | (BulkInData[10] << 16)
               | (BulkInData[9] << 8) | BulkInData[8]; 

//  AccR = (topro_read_reg(udev,0x83) << 24) | (topro_read_reg(udev,0x82) << 16)
//                 | (topro_read_reg(udev,0x81) << 8) | topro_read_reg(udev,0x80);


//  AccG = (topro_read_reg(udev,0x87) << 24) | (topro_read_reg(udev,0x86) << 16)
//                 | (topro_read_reg(udev,0x85) << 8) | topro_read_reg(udev,0x84);              

//  AccB = (topro_read_reg(udev,0x8b) << 24) | (topro_read_reg(udev,0x8a) << 16)
//                 | (topro_read_reg(udev,0x89) << 8) | topro_read_reg(udev,0x88);
  AccR = (AccR) / (640 * 480/4);
  AccG = (AccG) / (640 * 480/2);
  AccB = (AccB) / (640 * 480/4);

  if (AccR > 255 || AccR < 0 || AccG > 255 || AccG < 0 || AccB > 255 || AccB < 0){
    printk("AccRGB bulk_in data Error\n");
    return;
  }
  
  printk("AccRGB %d %d %d \n",(int)AccR,(int)AccG,(int)AccB);
		if (abs(AccR - AccG) > 2){
			if (AccR > AccG){
				if (AccR > AccG + 20)

					RFactor = max(0, RFactor - min(((int)AccR - (int)AccG), 4));
				else
					RFactor--;
			}
			else{
				if (AccR < AccG - 20)
					RFactor = min(100, RFactor + min((int)AccG, 4));
				else
					RFactor++;
			}
      
		}
		if (abs(AccB - AccG) > 2){
			if (AccB > AccG){
				if (AccB > AccG + 20)
					BFactor = max(0, BFactor - min(((int)AccB - (int)AccG), 6));

				else
					BFactor--;
			}
			else{
				if (AccB < AccG - 20)
					BFactor = min(100, BFactor + min(((int)AccG - (int)AccB), 6));
				else
					BFactor++;
			}

		}
 
 
/* 

   if (test==0){
     RFactor+=10;
     BFactor-=10;
     if (RFactor==100)
      test=1;
   }
   else{
     RFactor-=10;
     BFactor+=10;

     if (RFactor==0)
      test=0;
   }
     
*/   
 // printk("AutoWhiteBalance R %d,B %d\n",RFactor,BFactor);          

    topro_set_sensorgain(udev,GlobalGain);

//  topro_set_awb(udev,0x00,RFactor);
//  topro_set_awb(udev,0x02,BFactor);
    return;
}
/*
void WriteDefaultToFile()
{


  FILE *valuefile;
  valuefile = fopen("/usr/local/topro/topro.ini","r");
  if(valuefile == NULL){

    chdir("/usr/local");
    mkdir("topro",0);

    valuefile = fopen("/usr/local/topro/topro.ini","w");
    if (valuefile != NULL){
      fprintf(valuefile,"%d\n",GAMMA_R_GAIN);

      fprintf(valuefile,"%d\n",GAMMA_G_GAIN);
      fprintf(valuefile,"%d\n",GAMMA_B_GAIN);
      fprintf(valuefile,"%d\n",GAMMA_CONTRAST);
      fprintf(valuefile,"%d\n",GAMMA_BRIGHTNESS);

      fprintf(valuefile,"%d\n",DEFAULT_EXPOSURE);
      fclose(valuefile);
    }
  }
}
*/
////////////////////////////////////////////////////////////////////////////////////
// Sharp CCD control
////////////////////////////////////////////////////////////////////////////////////
void CCD_Control(struct usb_device *udev, unsigned char* pData, int Length)
{

//	unsigned char WriteBuf, ReadBuf;
	int i, j;
	unsigned char bData, bOut;
//	unsigned short int bAddr;
	unsigned char* pBuf;


	topro_write_reg(udev, GPIO_IO, 0x91);	//bit-6 LED, bit-5 Flip control, bit-3~0 4 bits for R/W EEPROM (bit-3,2,1 out, bit-0 in)
	pBuf = pData;				//sharp

	for (i=Length;i>0;i-=8){
		if (i > 8)
			j = 8;
		else
			j = i;
		bData = *pBuf;
//		pBuf--;		//sony
		pBuf++;		//sharp
		while(j>0){
			bOut = bData & 0x01 ? 0x02 : 0x00;
			topro_write_reg(udev, GPIO_DATA, bOut | 0x00);	//xxxx00ix

			topro_write_reg(udev, GPIO_DATA, bOut | 0x04);	//xxxx01ix
//      printk("Write GPIO_DATA %x %x\n",  bOut | 0x00,  bOut | 0x04);
			bData >>= 1;
			j--;
		}
	}
	topro_write_reg(udev, GPIO_DATA, 0x02);	//xxxx001x
	topro_write_reg(udev, GPIO_DATA, 0x0e);	//xxxx111x
//  sleep(10);

}

